# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Functions for testing silverkite estimators."""

import datetime

import pandas as pd

from greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper import cols_interact
from greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper import generate_holiday_events
from greykite.common.features.timeseries_features import convert_date_to_continuous_time


def params_components():
    """Parameters for ``forecast_silverkite``"""
    autoreg_dict = {
        "lag_dict": {"orders": [7]},
        "agg_lag_dict": {
            "orders_list": [[7, 7*2, 7*3]],
            "interval_list": [(7, 7*2)]},
        "series_na_fill_func": lambda s: s.bfill().ffill()}

    uncertainty_dict = {
        "uncertainty_method": "simple_conditional_residuals",
        "params": {
            "conditional_cols": ["dow"],
            "quantiles": [0.025, 0.975],
            "quantile_estimation_method": "normal_fit",
            "sample_size_thresh": 5,
            "small_sample_size_method": "std_quantiles",
            "small_sample_size_quantile": 0.98}}

    # generate holidays
    countries = ["US", "India"]
    holidays_to_model_separately = [
        "New Year's Day",
        "Christmas Day",
        "Independence Day",
        "Thanksgiving",
        "Labor Day",
        "Memorial Day",
        "Veterans Day"]
    event_df_dict = generate_holiday_events(
        countries=countries,
        holidays_to_model_separately=holidays_to_model_separately,
        year_start=2015,
        year_end=2025,
        pre_num=2,
        post_num=2)
    # constant event effect at daily level
    event_cols = [f"Q('events_{key}')" for key in event_df_dict.keys()]
    interaction_cols = cols_interact(
        static_col="is_weekend",
        fs_name="tow",
        fs_order=4,
        fs_seas_name="weekly")
    extra_pred_cols = ["ct_sqrt", "dow_hr", "ct1", "ct1:tod", "regressor1", "regressor2"] + \
        event_cols + interaction_cols

    # seasonality terms
    fs_components_df = pd.DataFrame({
        "name": ["tod", "tow", "ct1"],
        "period": [24.0, 7.0, 1.0],
        "order": [12, 4, 5],
        "seas_names": ["daily", "weekly", "yearly"]})

    # changepoints
    changepoints_dict = dict(
        method="custom",
        dates=["2018-01-01", "2019-01-02-16", "2019-01-03", "2019-02-01"],
        continuous_time_col="ct2")

    return {
        "coverage": 0.95,
        "origin_for_time_vars": convert_date_to_continuous_time(datetime.datetime(2018, 1, 3)),
        "extra_pred_cols": extra_pred_cols,
        "train_test_thresh": None,
        "training_fraction": None,
        "fit_algorithm": "ridge",
        "daily_event_df_dict": event_df_dict,
        "changepoints_dict": changepoints_dict,
        "fs_components_df": fs_components_df,
        "autoreg_dict": autoreg_dict,
        "min_admissible_value": None,
        "max_admissible_value": None,
        "uncertainty_dict": uncertainty_dict
    }



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# original author: Rachit Arora

from greykite.algo.forecast.silverkite.constants.silverkite_column import SilverkiteColumnMixin
from greykite.algo.forecast.silverkite.constants.silverkite_component import SilverkiteComponentsEnumMixin
from greykite.algo.forecast.silverkite.constants.silverkite_holiday import SilverkiteHolidayMixin
from greykite.algo.forecast.silverkite.constants.silverkite_seasonality import SilverkiteSeasonalityEnumMixin
from greykite.algo.forecast.silverkite.constants.silverkite_time_frequency import SilverkiteTimeFrequencyEnumMixin


class SilverkiteConstant(
    SilverkiteColumnMixin,
    SilverkiteComponentsEnumMixin,
    SilverkiteHolidayMixin,
    SilverkiteSeasonalityEnumMixin,
    SilverkiteTimeFrequencyEnumMixin
):
    """Uses the appropriate constant mixins to provide all the constants that will be used by Silverkite."""
    pass


default_silverkite_constant = SilverkiteConstant()



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Computation pipeline for end-to-end forecasting
using any ``BaseForecastEstimator``.
Runs cross validation, backtest, and forecast.
"""

import functools
import warnings
from dataclasses import dataclass

import pandas as pd
from sklearn import clone
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection import RandomizedSearchCV
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.evaluation import EvaluationMetricEnum
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import get_integer
from greykite.common.time_properties import min_gap_in_seconds
from greykite.framework.constants import COMPUTATION_N_JOBS
from greykite.framework.constants import CV_REPORT_METRICS_ALL
from greykite.framework.input.univariate_time_series import UnivariateTimeSeries
from greykite.framework.output.univariate_forecast import UnivariateForecast
from greykite.framework.pipeline.utils import get_basic_pipeline
from greykite.framework.pipeline.utils import get_default_time_parameters
from greykite.framework.pipeline.utils import get_forecast
from greykite.framework.pipeline.utils import get_hyperparameter_searcher
from greykite.sklearn.cross_validation import RollingTimeSeriesSplit
from greykite.sklearn.estimator.simple_silverkite_estimator import SimpleSilverkiteEstimator


@dataclass
class ForecastResult:
    """Forecast results. Contains results from cross-validation, backtest, and forecast,
    the trained model, and the original input data.
    """
    timeseries: UnivariateTimeSeries = None
    """Input time series in standard format with stats and convenient plot functions."""
    grid_search: RandomizedSearchCV = None
    """Result of cross-validation grid search on training dataset. The relevant attributes are:

        * ``cv_results_`` cross-validation scores
        * ``best_estimator_`` the model used for backtesting
        * ``best_params_`` the optimal parameters used for backtesting.

    Also see `~greykite.framework.utils.result_summary.summarize_grid_search_results`.
    We recommend using this function to extract results, rather than accessing ``cv_results_`` directly.
    """
    model: Pipeline = None
    """Model fitted on full dataset, using the best parameters selected via cross-validation.
    Has ``.fit()``, ``.predict()``, and diagnostic functions depending on the model.
    """
    backtest: UnivariateForecast = None
    """Forecast on backtest period.
    Backtest period is a holdout test set to check forecast quality against
    the most recent actual values available.
    The best model from cross validation is refit on data prior to this period.
    The timestamps in ``backtest.df`` are sorted in ascending order.
    Has a ``.plot()`` method and attributes to get
    forecast vs actuals, evaluation results.
    """
    forecast: UnivariateForecast = None
    """Forecast on future period.
    Future dates are after the train end date, following the holdout test set.
    The best model from cross validation is refit on data prior to this period.
    The timestamps in ``forecast.df`` are sorted in ascending order.
    Has a ``.plot()`` method and attributes to get
    forecast vs actuals, evaluation results.
    """


def validate_pipeline_input(pipeline_function):
    """Decorator that validates inputs to forecast_pipeline function and sets defaults"""
    @functools.wraps(pipeline_function)
    def pipeline_wrapper(
            # The arguments to this wrapper must be identical to forecast_pipeline() function.
            # We don't use **kwargs
            # because it's easier to check parameters directly.
            # input
            df: pd.DataFrame,
            time_col=TIME_COL,
            value_col=VALUE_COL,
            date_format=None,
            tz=None,
            freq=None,
            train_end_date=None,
            anomaly_info=None,
            # model
            pipeline=None,
            regressor_cols=None,
            estimator=SimpleSilverkiteEstimator(),
            hyperparameter_grid=None,
            hyperparameter_budget=None,
            n_jobs=COMPUTATION_N_JOBS,
            verbose=1,
            # forecast
            forecast_horizon=None,
            coverage=0.95,
            test_horizon=None,
            periods_between_train_test=None,
            agg_periods=None,
            agg_func=None,
            # evaluation
            score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
            score_func_greater_is_better=False,
            cv_report_metrics=None,
            null_model_params=None,
            relative_error_tolerance=None,
            # CV
            cv_horizon=None,
            cv_min_train_periods=None,
            cv_expanding_window=False,
            cv_use_most_recent_splits=False,
            cv_periods_between_splits=None,
            cv_periods_between_train_test=0,
            cv_max_splits=3):
        if coverage is not None and (coverage < 0 or coverage > 1):
            raise ValueError(f"coverage must be between 0 and 1, found {coverage}")
        if relative_error_tolerance is not None and relative_error_tolerance < 0:
            raise ValueError(f"relative_error_tolerance must non-negative, found {relative_error_tolerance}")

        # default values for forecast horizon, test, and cross-validation parameters
        period = min_gap_in_seconds(df=df, time_col=time_col)
        num_observations = df.shape[0]
        default_time_params = get_default_time_parameters(
            period=period,
            num_observations=num_observations,
            forecast_horizon=forecast_horizon,
            test_horizon=test_horizon,
            periods_between_train_test=periods_between_train_test,
            cv_horizon=cv_horizon,
            cv_min_train_periods=cv_min_train_periods,
            cv_periods_between_train_test=cv_periods_between_train_test)
        forecast_horizon = default_time_params.get("forecast_horizon")
        test_horizon = default_time_params.get("test_horizon")
        periods_between_train_test = default_time_params.get("periods_between_train_test")
        cv_horizon = default_time_params.get("cv_horizon")
        cv_min_train_periods = default_time_params.get("cv_min_train_periods")
        cv_periods_between_train_test = default_time_params.get("cv_periods_between_train_test")

        # ensures the values are integers in the proper domain
        if hyperparameter_budget is not None:
            hyperparameter_budget = get_integer(
                hyperparameter_budget,
                "hyperparameter_budget",
                min_value=1)

        if (cv_horizon == 0 or cv_max_splits == 0) and test_horizon == 0:
            raise ValueError("Either CV or backtest must be enabled."
                             " Set cv_horizon and cv_max_splits to nonzero values to enable CV."
                             " Set test_horizon to nonzero value to enable backtest."
                             " It's important to check model"
                             " performance on historical data.")

        if test_horizon == 0:
            warnings.warn("No data selected for test (test_horizon=0). "
                          "It is important to check out of sample performance")

        # checks horizon against data size
        if num_observations < forecast_horizon * 2:
            warnings.warn(f"Not enough training data to forecast the full forecast_horizon."
                          " Exercise extra caution with"
                          f" forecasted values after {num_observations // 2} periods.")

        if test_horizon > num_observations:
            raise ValueError(f"test_horizon ({test_horizon}) is too large."
                             " Must be less than the number "
                             f"of input data points: {num_observations})")

        if test_horizon > forecast_horizon:
            warnings.warn(f"test_horizon should never be larger than forecast_horizon.")

        if test_horizon > num_observations // 3:
            warnings.warn(f"test_horizon should be <= than 1/3 of the data set size to allow enough data to train"
                          f" a backtest model. Consider reducing to {num_observations // 3}. If this is smaller"
                          f" than the forecast_horizon, you will need to make a trade-off between setting"
                          f" test_horizon=forecast_horizon and having enough data left over to properly"
                          f" train a realistic backtest model.")

        log_message(f"forecast_horizon: {forecast_horizon}", LoggingLevelEnum.INFO)
        log_message(f"test_horizon: {test_horizon}", LoggingLevelEnum.INFO)
        log_message(f"cv_horizon: {cv_horizon}", LoggingLevelEnum.INFO)

        return pipeline_function(
            df,
            time_col=time_col,
            value_col=value_col,
            date_format=date_format,
            tz=tz,
            freq=freq,
            train_end_date=train_end_date,
            anomaly_info=anomaly_info,
            pipeline=pipeline,
            regressor_cols=regressor_cols,
            estimator=estimator,
            hyperparameter_grid=hyperparameter_grid,
            hyperparameter_budget=hyperparameter_budget,
            n_jobs=n_jobs,
            verbose=verbose,
            forecast_horizon=forecast_horizon,
            coverage=coverage,
            test_horizon=test_horizon,
            periods_between_train_test=periods_between_train_test,
            agg_periods=agg_periods,
            agg_func=agg_func,
            score_func=score_func,
            score_func_greater_is_better=score_func_greater_is_better,
            cv_report_metrics=cv_report_metrics,
            null_model_params=null_model_params,
            relative_error_tolerance=relative_error_tolerance,
            cv_horizon=cv_horizon,
            cv_min_train_periods=cv_min_train_periods,
            cv_expanding_window=cv_expanding_window,
            cv_use_most_recent_splits=cv_use_most_recent_splits,
            cv_periods_between_splits=cv_periods_between_splits,
            cv_periods_between_train_test=cv_periods_between_train_test,
            cv_max_splits=cv_max_splits
        )
    return pipeline_wrapper


@validate_pipeline_input
def forecast_pipeline(
        # input
        df: pd.DataFrame,
        time_col=TIME_COL,
        value_col=VALUE_COL,
        date_format=None,
        tz=None,
        freq=None,
        train_end_date=None,
        anomaly_info=None,
        # model
        pipeline=None,
        regressor_cols=None,
        estimator=SimpleSilverkiteEstimator(),
        hyperparameter_grid=None,
        hyperparameter_budget=None,
        n_jobs=COMPUTATION_N_JOBS,
        verbose=1,
        # forecast
        forecast_horizon=None,
        coverage=0.95,
        test_horizon=None,
        periods_between_train_test=None,
        agg_periods=None,
        agg_func=None,
        # evaluation
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        score_func_greater_is_better=False,
        cv_report_metrics=CV_REPORT_METRICS_ALL,
        null_model_params=None,
        relative_error_tolerance=None,
        # CV
        cv_horizon=None,
        cv_min_train_periods=None,
        cv_expanding_window=False,
        cv_use_most_recent_splits=False,
        cv_periods_between_splits=None,
        cv_periods_between_train_test=None,
        cv_max_splits=3):
    """Computation pipeline for end-to-end forecasting.

    Trains a forecast model end-to-end:

        1. checks input data
        2. runs cross-validation to select optimal hyperparameters e.g. best model
        3. evaluates best model on test set
        4. provides forecast of best model (re-trained on all data) into the future

    Returns forecasts with methods to plot and see diagnostics.
    Also returns the fitted pipeline and CV results.

    Provides a high degree of customization over training and evaluation parameters:

        1. model
        2. cross validation
        3. evaluation
        4. forecast horizon

    See test cases for examples.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Timeseries data to forecast.
        Contains columns [`time_col`, `value_col`], and optional regressor columns
        Regressor columns should include future values for prediction

    time_col : `str`, default TIME_COL in constants.py
        name of timestamp column in df

    value_col : `str`, default VALUE_COL in constants.py
        name of value column in df (the values to forecast)

    date_format : `str` or None, default None
        strftime format to parse time column, eg ``%m/%d/%Y``.
        Note that ``%f`` will parse all the way up to nanoseconds.
        If None (recommended), inferred by `pandas.to_datetime`.

    tz : `str` or None, default None
        Passed to `pandas.tz_localize` to localize the timestamp

    freq : `str` or None, default None
        Frequency of input data. Used to generate future dates for prediction.
        Frequency strings can have multiples, e.g. '5H'.
        See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        for a list of frequency aliases.
        If None, inferred by `pandas.infer_freq`.
        Provide this parameter if ``df`` has missing timepoints.

    train_end_date : `datetime.datetime`, optional, default None
        Last date to use for fitting the model. Forecasts are generated after this date.
        If None, it is set to the last date with a non-null value in
        ``value_col`` of ``df``.

    anomaly_info : `dict` or `list` [`dict`] or None, default None
        Anomaly adjustment info. Anomalies in ``df``
        are corrected before any forecasting is done.

        If None, no adjustments are made.

        A dictionary containing the parameters to
        `~greykite.common.features.adjust_anomalous_data.adjust_anomalous_data`.
        See that function for details.
        The possible keys are:

            ``"value_col"`` : `str`
                The name of the column in ``df`` to adjust. You may adjust the value
                to forecast as well as any numeric regressors.
            ``"anomaly_df"`` : `pandas.DataFrame`
                Adjustments to correct the anomalies.
            ``"start_date_col"``: `str`, default START_DATE_COL
                Start date column in ``anomaly_df``.
            ``"end_date_col"``: `str`, default END_DATE_COL
                End date column in ``anomaly_df``.
            ``"adjustment_delta_col"``: `str` or None, default None
                Impact column in ``anomaly_df``.
            ``"filter_by_dict"``: `dict` or None, default None
                Used to filter ``anomaly_df`` to the relevant anomalies for
                the ``value_col`` in this dictionary.
                Key specifies the column name, value specifies the filter value.
            ``"filter_by_value_col""``: `str` or None, default None
                Adds ``{filter_by_value_col: value_col}`` to ``filter_by_dict``
                if not None, for the ``value_col`` in this dictionary.
            ``"adjustment_method"`` : `str` ("add" or "subtract"), default "add"
                How to make the adjustment, if ``adjustment_delta_col`` is provided.

        Accepts a list of such dictionaries to adjust multiple columns in ``df``.

    pipeline : `sklearn.pipeline.Pipeline` or None, default None
        Pipeline to fit. The final named step must be called "estimator".
        If None, will use the default Pipeline from
        `~greykite.framework.pipeline.utils.get_basic_pipeline`.

    regressor_cols : `list` [`str`] or None, default None
        A list of regressor columns used in the training and prediction DataFrames.
        It should contain only the regressors that are being used in the grid search.
        If None, no regressor columns are used.
        Regressor columns that are unavailable in ``df`` are dropped.

    estimator : instance of an estimator that implements `greykite.algo.models.base_forecast_estimator.BaseForecastEstimator`
        Estimator to use as the final step in the pipeline.
        Ignored if ``pipeline`` is provided.

    forecast_horizon : `int` or None, default None
        Number of periods to forecast into the future. Must be > 0.
        If None, default is determined from input data frequency

    coverage : `float` or None, default=0.95
        Intended coverage of the prediction bands (0.0 to 1.0)
        If None, the upper/lower predictions are not returned
        Ignored if `pipeline` is provided. Uses coverage of the ``pipeline`` estimator instead.

    test_horizon : `int` or None, default None
        Numbers of periods held back from end of df for test.
        The rest is used for cross validation.
        If None, default is forecast_horizon. Set to 0 to skip backtest.

    periods_between_train_test : `int` or None, default None
        Number of periods for the gap between train and test data.
        If None, default is 0.

    agg_periods : `int` or None, default None
        Number of periods to aggregate before evaluation.

        Model is fit and forecasted on the dataset's original frequency.

        Before evaluation, the actual and forecasted values are aggregated,
        using rolling windows of size ``agg_periods`` and the function
        ``agg_func``. (e.g. if the dataset is hourly, use ``agg_periods=24, agg_func=np.sum``,
        to evaluate performance on the daily totals).

        If None, does not aggregate before evaluation.

        Currently, this is only used when calculating CV metrics and
        the R2_null_model_score metric in backtest/forecast. No pre-aggregation
        is applied for the other backtest/forecast evaluation metrics.

    agg_func : callable or None, default None
        Takes an array and returns a number, e.g. np.max, np.sum.

        Defines how to aggregate rolling windows of actual and predicted values
        before evaluation.

        Ignored if ``agg_periods`` is None.

        Currently, this is only used when calculating CV metrics and
        the R2_null_model_score metric in backtest/forecast. No pre-aggregation
        is applied for the other backtest/forecast evaluation metrics.

    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function used to select optimal model in CV.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.

    score_func_greater_is_better : `bool`, default False
        True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        Must be provided if ``score_func`` is a callable (custom function).
        Ignored if ``score_func`` is a string, because the direction is known.

    cv_report_metrics : `str`, or `list` [`str`], or None, default `~greykite.common.constants.CV_REPORT_METRICS_ALL`
        Additional metrics to compute during CV, besides the one specified by ``score_func``.

            - If the string constant `greykite.framework.constants.CV_REPORT_METRICS_ALL`,
              computes all metrics in ``EvaluationMetricEnum``. Also computes
              ``FRACTION_OUTSIDE_TOLERANCE`` if ``relative_error_tolerance`` is not None.
              The results are reported by the short name (``.get_metric_name()``) for ``EvaluationMetricEnum``
              members and ``FRACTION_OUTSIDE_TOLERANCE_NAME`` for ``FRACTION_OUTSIDE_TOLERANCE``.
              These names appear in the keys of ``forecast_result.grid_search.cv_results_``
              returned by this function.
            - If a list of strings, each of the listed metrics is computed. Valid strings are
              `~greykite.common.evaluation.EvaluationMetricEnum` member names
              and `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.

              For example::

                ["MeanSquaredError", "MeanAbsoluteError", "MeanAbsolutePercentError", "MedianAbsolutePercentError", "FractionOutsideTolerance2"]

            - If None, no additional metrics are computed.

    null_model_params : `dict` or None, default None
        Defines baseline model to compute ``R2_null_model_score`` evaluation metric.
        ``R2_null_model_score`` is the improvement in the loss function relative
        to a null model. It can be used to evaluate model quality with respect to
        a simple baseline. For details, see
        `~greykite.common.evaluation.r2_null_model_score`.

        The null model is a `~sklearn.dummy.DummyRegressor`,
        which returns constant predictions.

        Valid keys are "strategy", "constant", "quantile".
        See `~sklearn.dummy.DummyRegressor`. For example::

            null_model_params = {
                "strategy": "mean",
            }
            null_model_params = {
                "strategy": "median",
            }
            null_model_params = {
                "strategy": "quantile",
                "quantile": 0.8,
            }
            null_model_params = {
                "strategy": "constant",
                "constant": 2.0,
            }

        If None, ``R2_null_model_score`` is not calculated.

        Note: CV model selection always optimizes ``score_func`, not
        the ``R2_null_model_score``.

    relative_error_tolerance : `float` or None, default None
        Threshold to compute the ``Outside Tolerance`` metric,
        defined as the fraction of forecasted values whose relative
        error is strictly greater than ``relative_error_tolerance``.
        For example, 0.05 allows for 5% relative error.
        If `None`, the metric is not computed.

    hyperparameter_grid : `dict`, `list` [`dict`] or None, default None
        Sets properties of the steps in the pipeline,
        and specifies combinations to search over.
        Should be valid input to `sklearn.model_selection.GridSearchCV` (param_grid)
        or `sklearn.model_selection.RandomizedSearchCV` (param_distributions).

        Prefix transform/estimator attributes by the name of the step in the pipeline.
        See details at: https://scikit-learn.org/stable/modules/compose.html#nested-parameters

        If None, uses the default pipeline parameters.

    hyperparameter_budget : `int` or None, default None
        Max number of hyperparameter sets to try within the ``hyperparameter_grid`` search space

        Runs a full grid search if ``hyperparameter_budget`` is sufficient to exhaust full
        ``hyperparameter_grid``, otherwise samples uniformly at random from the space.

        If None, uses defaults:

            * full grid search if all values are constant
            * 10 if any value is a distribution to sample from

    n_jobs : `int` or None, default `~greykite.framework.constants.COMPUTATION_N_JOBS`
        Number of jobs to run in parallel
        (the maximum number of concurrently running workers).
        ``-1`` uses all CPUs. ``-2`` uses all CPUs but one.
        ``None`` is treated as 1 unless in a `joblib.Parallel` backend context
        that specifies otherwise.

    verbose : `int`, default 1
        Verbosity level during CV.
        if > 0, prints number of fits
        if > 1, prints fit parameters, total score + fit time
        if > 2, prints train/test scores

    cv_horizon : `int` or None, default None
        Number of periods in each CV test set
        If None, default is ``forecast_horizon``.
        Set either ``cv_horizon`` or ``cv_max_splits`` to 0 to skip CV.

    cv_min_train_periods : `int` or None, default None
        Minimum number of periods for training each CV fold.
        If cv_expanding_window is False, every training period is this size
        If None, default is 2 * ``cv_horizon``

    cv_expanding_window : `bool`, default False
        If True, training window for each CV split is fixed to the first available date.
        Otherwise, train start date is sliding, determined by ``cv_min_train_periods``.

    cv_use_most_recent_splits: `bool`, default False
        If True, splits from the end of the dataset are used.
        Else a sampling strategy is applied. Check
        `~greykite.sklearn.cross_validation.RollingTimeSeriesSplit._sample_splits`
        for details.

    cv_periods_between_splits : `int` or None, default None
        Number of periods to slide the test window between CV splits
        If None, default is ``cv_horizon``

    cv_periods_between_train_test : `int` or None, default None
        Number of periods for the gap between train and test in a CV split.
        If None, default is ``periods_between_train_test``.

    cv_max_splits : `int` or None, default 3
        Maximum number of CV splits.
        Given the above configuration, samples up to max_splits train/test splits,
        preferring splits toward the end of available data. If None, uses all splits.
        Set either ``cv_horizon`` or ``cv_max_splits`` to 0 to skip CV.

    Returns
    -------
    forecast_result : :class:`~greykite.framework.pipeline.pipeline.ForecastResult`
        Forecast result. See :class:`~greykite.framework.pipeline.pipeline.ForecastResult`
        for details.

            * If ``cv_horizon=0``, ``forecast_result.grid_search.best_estimator_``
              and ``forecast_result.grid_search.best_params_`` attributes are defined
              according to the provided single set of parameters. There must be a single
              set of parameters to skip cross-validation.
            * If ``test_horizon=0``, ``forecast_result.backtest`` is None.
    """
    if hyperparameter_grid is None:
        hyperparameter_grid = {}
    # Loads full dataset
    ts = UnivariateTimeSeries()
    ts.load_data(
        df=df,
        time_col=time_col,
        value_col=value_col,
        freq=freq,
        date_format=date_format,
        tz=tz,
        train_end_date=train_end_date,
        regressor_cols=regressor_cols,
        anomaly_info=anomaly_info)

    # Splits data into training and test sets. ts.df uses standardized column names
    if test_horizon == 0:
        train_df = ts.fit_df
        train_y = ts.fit_y
        test_df = pd.DataFrame(columns=list(df.columns))
    else:
        # Make sure to refit best_pipeline appropriately
        train_df, test_df, train_y, test_y = train_test_split(
            ts.fit_df,
            ts.fit_y,
            train_size=ts.fit_df.shape[0] - test_horizon - periods_between_train_test,
            test_size=test_horizon + periods_between_train_test,
            shuffle=False)  # this is important since this is timeseries forecasting!
    log_message(f"Train size: {train_df.shape[0]}. Test size: {test_df.shape[0]}", LoggingLevelEnum.INFO)

    # Defines default training pipeline
    if pipeline is None:
        pipeline = get_basic_pipeline(
            estimator=estimator,
            score_func=score_func,
            score_func_greater_is_better=score_func_greater_is_better,
            agg_periods=agg_periods,
            agg_func=agg_func,
            relative_error_tolerance=relative_error_tolerance,
            coverage=coverage,
            null_model_params=null_model_params,
            regressor_cols=ts.regressor_cols)

    # Searches for the best parameters, and refits model with selected parameters on the entire training set
    if cv_horizon == 0 or cv_max_splits == 0:
        # No cross-validation. Only one set of hyperparameters is allowed.
        try:
            if len(ParameterGrid(hyperparameter_grid)) > 1:
                raise ValueError(
                    "CV is required to identify the best model because there are multiple options "
                    "in `hyperparameter_grid`. Either provide a single option or set `cv_horizon` and `cv_max_splits` "
                    "to nonzero values.")
        except TypeError:  # Parameter value is not iterable
            raise ValueError(
                "CV is required to identify the best model because `hyperparameter_grid` contains "
                "a distribution. Either remove the distribution or set `cv_horizon` and `cv_max_splits` "
                "to nonzero values.")

        # Fits model to entire train set. Params must be set manually since it's not done by grid search
        params = {k: v[0] for k, v in hyperparameter_grid.items()}  # unpack lists, `v` is a singleton list with the parameter value
        best_estimator = pipeline.set_params(**params).fit(train_df, train_y)

        # Wraps this model in a dummy RandomizedSearchCV object to return the backtest model
        grid_search = get_hyperparameter_searcher(
            hyperparameter_grid=hyperparameter_grid,
            model=pipeline,
            cv=None,  # no cross-validation
            hyperparameter_budget=hyperparameter_budget,
            n_jobs=n_jobs,
            verbose=verbose,
            score_func=score_func,
            score_func_greater_is_better=score_func_greater_is_better,
            cv_report_metrics=cv_report_metrics,
            agg_periods=agg_periods,
            agg_func=agg_func,
            relative_error_tolerance=relative_error_tolerance)
        # Sets relevant attributes. Others are undefined (cv_results_, best_score_, best_index_, scorer_, refit_time_)
        grid_search.best_estimator_ = best_estimator
        grid_search.best_params_ = params
        grid_search.n_splits_ = 0
    else:
        # Defines cross-validation splitter
        cv = RollingTimeSeriesSplit(
            forecast_horizon=cv_horizon,
            min_train_periods=cv_min_train_periods,
            expanding_window=cv_expanding_window,
            use_most_recent_splits=cv_use_most_recent_splits,
            periods_between_splits=cv_periods_between_splits,
            periods_between_train_test=cv_periods_between_train_test,
            max_splits=cv_max_splits)

        # Defines grid search approach for CV
        grid_search = get_hyperparameter_searcher(
            hyperparameter_grid=hyperparameter_grid,
            model=pipeline,
            cv=cv,
            hyperparameter_budget=hyperparameter_budget,
            n_jobs=n_jobs,
            verbose=verbose,
            score_func=score_func,
            score_func_greater_is_better=score_func_greater_is_better,
            cv_report_metrics=cv_report_metrics,
            agg_periods=agg_periods,
            agg_func=agg_func,
            relative_error_tolerance=relative_error_tolerance)
        grid_search.fit(train_df, train_y)
        best_estimator = grid_search.best_estimator_

    # Evaluates historical performance, fits model to all data (train+test)
    if test_horizon > 0:
        backtest_train_end_date = train_df[TIME_COL].max()
        # Uses pd.date_range because pd.Timedelta does not work for complicated frequencies e.g. "W-MON"
        backtest_test_start_date = pd.date_range(
            start=backtest_train_end_date,
            periods=periods_between_train_test + 2,  # Adds 2 as start parameter is inclusive
            freq=ts.freq)[-1]
        backtest = get_forecast(
            df=ts.fit_df,  # Backtest needs to happen on fit_df, not on the entire df
            trained_model=best_estimator,
            train_end_date=backtest_train_end_date,
            test_start_date=backtest_test_start_date,
            forecast_horizon=test_horizon,
            xlabel=time_col,
            ylabel=value_col,
            relative_error_tolerance=relative_error_tolerance)
        best_pipeline = clone(best_estimator)  # Copies optimal parameters
        best_pipeline.fit(ts.fit_df, ts.y)  # Refits this model on entire training dataset
    else:
        backtest = None  # Backtest training metrics are the same as forecast training metrics
        best_pipeline = best_estimator  # best_model is already fit to all data

    # Makes future predictions
    periods = forecast_horizon + periods_between_train_test
    future_df = ts.make_future_dataframe(
        periods=periods,
        include_history=True)

    forecast_train_end_date = ts.train_end_date
    # Uses pd.date_range because pd.Timedelta does not work for complicated frequencies e.g. "W-MON"
    forecast_test_start_date = pd.date_range(
        start=forecast_train_end_date,
        periods=periods_between_train_test + 2,  # Adds 2 as start parameter is inclusive
        freq=ts.freq)[-1]
    forecast = get_forecast(
        df=future_df,
        trained_model=best_pipeline,
        train_end_date=forecast_train_end_date,
        test_start_date=forecast_test_start_date,
        forecast_horizon=forecast_horizon,
        xlabel=time_col,
        ylabel=value_col,
        relative_error_tolerance=relative_error_tolerance)

    result = ForecastResult(
        timeseries=ts,
        grid_search=grid_search,
        model=best_pipeline,
        backtest=backtest,
        forecast=forecast
    )
    return result



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Input timeseries."""

import warnings
from datetime import datetime
from functools import partial
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd

from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.time_properties import describe_timeseries
from greykite.common.time_properties import get_canonical_data
from greykite.common.viz.timeseries_plotting import add_groupby_column
from greykite.common.viz.timeseries_plotting import flexible_grouping_evaluation
from greykite.common.viz.timeseries_plotting import grouping_evaluation
from greykite.common.viz.timeseries_plotting import plot_multivariate
from greykite.common.viz.timeseries_plotting import plot_univariate
from greykite.framework.constants import MEAN_COL_GROUP
from greykite.framework.constants import OVERLAY_COL_GROUP
from greykite.framework.constants import QUANTILE_COL_GROUP


class UnivariateTimeSeries:
    """Defines univariate time series input. The dataset can include regressors,
    but only one metric is designated as the target metric to forecast.

    Loads time series into a standard format. Provides statistics, plotting
    functions, and ability to generate future dataframe for prediction.

    Attributes
    ----------
    df: `pandas.DataFrame`
        Data frame containing timestamp and value, with standardized column names for internal use
        (TIME_COL, VALUE_COL). Rows are sorted by time index, and missing gaps between dates are filled
        in so that dates are spaced at regular intervals. Values are adjusted for anomalies
        according to ``anomaly_info``.
        The index can be timezone aware (but TIME_COL is not).
    y: `pandas.Series`, dtype float64
        Value of time series to forecast.
    time_stats: `dict`
        Summary statistics about the timestamp column.
    value_stats: `dict`
        Summary statistics about the value column.
    original_time_col: `str`
        Name of time column in original input data.
    original_value_col: `str`
        Name of value column in original input data.
    regressor_cols: `list` [`str`]
        A list of regressor columns in the training and prediction DataFrames.
    last_date_for_val: `datetime.datetime` or None, default None
        Date or timestamp corresponding  to last non-null value in ``df[original_value_col]``.
    last_date_for_reg: `datetime.datetime` or None, default None
        Date or timestamp corresponding to last non-null value in ``df[regressor_cols]``.
        If ``regressor_cols`` is None, ``last_date_for_reg`` is None.
    train_end_date: `datetime.datetime`
        Last date or timestamp in ``fit_df``. It is always less than or equal to
        minimum non-null values of ``last_date_for_val`` and ``last_date_for_reg``.
    fit_cols: `list` [`str`]
        A list of columns used in the training and prediction DataFrames.
    fit_df: `pandas.DataFrame`
        Data frame containing timestamp and value, with standardized column names for internal use.
        Will be used for fitting (train, cv, backtest).
    fit_y: `pandas.Series`, dtype float64
        Value of time series for fit_df.
    freq: `str`
        timeseries frequency, DateOffset alias, e.g. {'T' (minute), 'H', D', 'W', 'M' (month end), 'MS' (month start),
        'Y' (year end), 'Y' (year start)}
        See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    anomaly_info : `dict` or `list` [`dict`] or None, default None
        Anomaly adjustment info. Anomalies in ``df``
        are corrected before any forecasting is done.
        See ``self.load_data()``
    df_before_adjustment : `pandas.DataFrame` or None, default None
        ``self.df`` before adjustment by ``anomaly_info``.
        Used by ``self.plot()`` to show the adjustment.
    """
    def __init__(self) -> None:
        self.df: Optional[pd.DataFrame] = None
        self.y: Optional[pd.Series] = None
        self.time_stats: Optional[Dict] = None
        self.value_stats: Optional[Dict] = None
        self.original_time_col: Optional[str] = None
        self.original_value_col: Optional[str] = None
        self.regressor_cols: List[str] = []
        self.last_date_for_val: Optional[datetime] = None
        self.last_date_for_reg: Optional[datetime] = None
        self.train_end_date: Optional[str] = None
        self.fit_cols: List[str] = []
        self.fit_df: Optional[pd.DataFrame] = None
        self.fit_y: Optional[pd.DataFrame] = None
        self.freq: Optional[str] = None
        self.anomaly_info: Optional[Union[Dict, List[Dict]]] = None
        self.df_before_adjustment: Optional[pd.DataFrame] = None

    def load_data(
            self,
            df: pd.DataFrame,
            time_col: str = TIME_COL,
            value_col: str = VALUE_COL,
            freq: str = None,
            date_format: str = None,
            tz: str = None,
            train_end_date: datetime = None,
            regressor_cols: List[str] = None,
            anomaly_info: Optional[Union[Dict, List[Dict]]] = None):
        """Loads data to internal representation. Parses date column,
        sets timezone aware index.
        Checks for irregularities and raises an error if input is invalid.
        Adjusts for anomalies according to ``anomaly_info``.

        Parameters
        ----------
        df : `pandas.DataFrame`
            Input timeseries. A data frame which includes the timestamp column
            as well as the value column.
        time_col : `str`
            The column name in ``df`` representing time for the time series data.
            The time column can be anything that can be parsed by pandas DatetimeIndex.
        value_col: `str`
            The column name which has the value of interest to be forecasted.
        freq : `str`, optional, default None
            Timeseries frequency, DateOffset alias, If None automatically inferred.
        date_format : `str`, optional, default None
            strftime format to parse time column, eg ``%m/%d/%Y``.
            Note that ``%f`` will parse all the way up to nanoseconds.
            If None (recommended), inferred by `pandas.to_datetime`.
        tz : `str` or pytz.timezone object, optional, default None
            Passed to `pandas.tz_localize` to localize the timestamp.
        train_end_date : `datetime.datetime`, optional, default None
            Last date to use for fitting the model. Forecasts are generated after this date.
            If None, it is set to the minimum of ``self.last_date_for_val`` and
            ``self.last_date_for_reg``.
        regressor_cols: `list` [`str`], optional, default None
            A list of regressor columns used in the training and prediction DataFrames.
            If None, no regressor columns are used.
            Regressor columns that are unavailable in ``df`` are dropped.
            anomaly_info : `dict` or None, default None
        anomaly_info : `dict` or `list` [`dict`] or None, default None
            Anomaly adjustment info. Anomalies in ``df``
            are corrected before any forecasting is done.

            If None, no adjustments are made.

            A dictionary containing the parameters to
            `~greykite.common.features.adjust_anomalous_data.adjust_anomalous_data`.
            See that function for details.
            The possible keys are:

                ``"value_col"`` : `str`
                    The name of the column in ``df`` to adjust. You may adjust the value
                    to forecast as well as any numeric regressors.
                ``"anomaly_df"`` : `pandas.DataFrame`
                    Adjustments to correct the anomalies.
                ``"start_date_col"``: `str`, default START_DATE_COL
                    Start date column in ``anomaly_df``.
                ``"end_date_col"``: `str`, default END_DATE_COL
                    End date column in ``anomaly_df``.
                ``"adjustment_delta_col"``: `str` or None, default None
                    Impact column in ``anomaly_df``.
                ``"filter_by_dict"``: `dict` or None, default None
                    Used to filter ``anomaly_df`` to the relevant anomalies for
                    the ``value_col`` in this dictionary.
                    Key specifies the column name, value specifies the filter value.
                ``"filter_by_value_col""``: `str` or None, default None
                    Adds ``{filter_by_value_col: value_col}`` to ``filter_by_dict``
                    if not None, for the ``value_col`` in this dictionary.
                ``"adjustment_method"`` : `str` ("add" or "subtract"), default "add"
                    How to make the adjustment, if ``adjustment_delta_col`` is provided.

            Accepts a list of such dictionaries to adjust multiple columns in ``df``.

        Returns
        -------
        self : Returns self.
            Sets ``self.df`` with standard column names,
            value adjusted for anomalies, and time gaps filled in,
            sorted by time index.
        """
        self.original_time_col = time_col
        self.original_value_col = value_col
        self.anomaly_info = anomaly_info

        canonical_data_dict = get_canonical_data(
            df=df,
            time_col=time_col,
            value_col=value_col,
            freq=freq,
            date_format=date_format,
            tz=tz,
            train_end_date=train_end_date,
            regressor_cols=regressor_cols,
            anomaly_info=anomaly_info)
        self.df = canonical_data_dict["df"]
        self.df_before_adjustment = canonical_data_dict["df_before_adjustment"]
        self.fit_df = canonical_data_dict["fit_df"]
        self.freq = canonical_data_dict["freq"]
        self.time_stats = canonical_data_dict["time_stats"]
        self.regressor_cols = canonical_data_dict["regressor_cols"]
        self.fit_cols = canonical_data_dict["fit_cols"]
        self.train_end_date = canonical_data_dict["train_end_date"]
        self.last_date_for_val = canonical_data_dict["last_date_for_val"]
        self.last_date_for_reg = canonical_data_dict["last_date_for_reg"]

        # y (possibly with null values) after gaps have been filled in and anomalies corrected
        self.y = self.df[VALUE_COL]
        self.fit_y = self.fit_df[VALUE_COL]

        # computes statistics of processed dataset
        self.describe_time_col()
        self.describe_value_col()  # compute value statistics

        log_message(f"last date for fit: {self.train_end_date}", LoggingLevelEnum.INFO)
        log_message(f"last date for {self.original_value_col}: {self.last_date_for_val}", LoggingLevelEnum.INFO)
        log_message(f"last date with any regressor: {self.last_date_for_reg}", LoggingLevelEnum.INFO)
        log_message(f"columns available to use as regressors: {', '.join(self.regressor_cols)}", LoggingLevelEnum.INFO)

        return self

    def describe_time_col(self):
        """Basic descriptive stats on the timeseries time column.

        Returns
        -------
        time_stats: `dict`
            Dictionary with descriptive stats on the timeseries time column.

                * data_points: int
                    number of time points
                * mean_increment_secs: float
                    mean frequency
                * min_timestamp: datetime64
                    start date
                * max_timestamp: datetime64
                    end date
        """
        if self.df is None:
            raise RuntimeError("Must load data before describing dataset")

        timeseries_info = describe_timeseries(df=self.df, time_col=TIME_COL)
        data_points = self.df.shape[0]
        mean_increment_secs = timeseries_info["mean_increment_secs"]
        min_timestamp = timeseries_info["min_timestamp"]
        max_timestamp = timeseries_info["max_timestamp"]

        log_message("Input time stats:", LoggingLevelEnum.INFO)
        log_message(f"  data points: {data_points}", LoggingLevelEnum.INFO)
        log_message(f"  avg increment (sec): {mean_increment_secs:.2f}", LoggingLevelEnum.INFO)
        log_message(f"  start date: {min_timestamp}", LoggingLevelEnum.INFO)
        log_message(f"  end date: {max_timestamp}", LoggingLevelEnum.INFO)

        time_stats = {
            "data_points": data_points,  # total number of time points, including missing ones
            "mean_increment_secs": mean_increment_secs,  # after filling in gaps
            "min_timestamp": min_timestamp,
            "max_timestamp": max_timestamp,
        }
        self.time_stats.update(time_stats)  # compute time statistics
        return time_stats

    def describe_value_col(self):
        """Basic descriptive stats on the timeseries value column.

        Returns
        -------
        value_stats : `dict` [`str`, `float`]
            Dict with keys: count, mean, std, min, 25%, 50%, 75%, max
        """
        if self.df is None:
            raise RuntimeError("Must load data before describing values")
        self.value_stats = self.df[VALUE_COL].describe()  # count is the total number of provided timepoints
        log_message("Input value stats:", LoggingLevelEnum.INFO)
        log_message(repr(self.value_stats), LoggingLevelEnum.INFO)
        return self.value_stats

    def make_future_dataframe(self, periods: int = None, include_history=True):
        """Extends the input data for prediction into the future.

        Includes the historical values (VALUE_COL) so this can be fed
        into a Pipeline that transforms input data for fitting, and for
        use in evaluation.

        Parameters
        ----------
        periods : int or None
            Number of periods to forecast.
            If there are no regressors, default is 30.
            If there are regressors, default is to predict all available dates.
        include_history : bool
            Whether to return historical dates and values with future dates.

        Returns
        -------
        future_df : `pandas.DataFrame`
            Dataframe with future timestamps for prediction.
            Contains columns for:

                * prediction dates (``TIME_COL``),
                * values (``VALUE_COL``),
                * optional regressors
        """
        if self.df is None:
            raise RuntimeError("Must load data before generating future dates.")

        # determines the number of future periods to predict
        if self.regressor_cols:
            max_regressor_periods = len(self.df[
                (self.df[TIME_COL] > self.train_end_date)
                & (self.df[TIME_COL] <= self.last_date_for_reg)
            ])
            if periods is None:
                periods = max_regressor_periods
            elif periods > max_regressor_periods:
                warnings.warn(
                    f"Provided periods '{periods}' is more than allowed ('{max_regressor_periods}') due to "
                    f"the length of regressor columns. Using '{max_regressor_periods}'.",
                    UserWarning)
                periods = max_regressor_periods
        elif periods is None:
            periods = 30

        # the future dates for prediction
        dates = pd.date_range(
            start=self.train_end_date,
            periods=periods + 1,  # an extra in case we include start
            freq=self.freq)
        dates = dates[dates > self.train_end_date]  # drops values up to train_end_date
        dates = dates[:periods]  # returns the correct number of periods

        if self.regressor_cols:
            # return TIME_COL, VALUE_COL, and regressors
            last_date_for_predict = dates.max()
            if include_history:
                valid_indices = (self.df[TIME_COL] <= last_date_for_predict)
            else:
                valid_indices = ((self.df[TIME_COL] > self.train_end_date)
                                 & (self.df[TIME_COL] <= last_date_for_predict))
            future_df = self.df[valid_indices]
        else:
            # return TIME_COL, VALUE_COL
            future_df = self.df.reindex(index=dates)
            future_df[TIME_COL] = future_df.index
            if include_history:
                future_df = pd.concat([self.fit_df, future_df], axis=0, sort=False)

        return future_df[self.fit_cols]

    def plot(
            self,
            color="rgb(32, 149, 212)",
            show_anomaly_adjustment=False,
            **kwargs):
        """Returns interactive plotly graph of the value against time.

        If anomaly info is provided, there is an option to show the anomaly adjustment.

        Parameters
        ----------
        color : `str`, default "rgb(32, 149, 212)" (light blue)
            Color of the value line (after adjustment, if applicable).
        show_anomaly_adjustment : `bool`, default False
            Whether to show the anomaly adjustment.
        kwargs : additional parameters
            Additional parameters to pass to
            `~greykite.common.viz.timeseries_plotting.plot_univariate`
            such as title and color.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            Interactive plotly graph of the value against time.

            See `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
            return value for how to plot the figure and add customization.
        """
        df = self.df.copy()
        # Plots value after anomaly adjustment
        y_col_style_dict = {
            VALUE_COL: dict(
                name=self.original_value_col,
                mode="lines",
                line=dict(
                    color=color,
                ),
                opacity=0.8
            )
        }
        if show_anomaly_adjustment:
            if self.anomaly_info is not None:
                # Adds value before adjustment to ``df``
                postfix = "_unadjusted"
                df[f"{VALUE_COL}{postfix}"] = self.df_before_adjustment[VALUE_COL]
                y_col_style_dict[f"{VALUE_COL}{postfix}"] = dict(
                    name=f"{self.original_value_col}{postfix}",
                    mode="lines",
                    line=dict(
                        color="#B3B3B3",  # light gray
                    ),
                    opacity=0.8
                )
            else:
                raise ValueError("There is no `anomaly_info` to show. `show_anomaly_adjustment` must be False.")
        return plot_multivariate(
            df,
            TIME_COL,
            y_col_style_dict,
            xlabel=self.original_time_col,
            ylabel=self.original_value_col,
            **kwargs)

    def get_grouping_evaluation(
            self,
            aggregation_func=np.nanmean,
            aggregation_func_name="mean",
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None):
        """Group-wise computation of aggregated timeSeries value.
        Can be used to evaluate error/ aggregated value by a time feature,
        over time, or by a user-provided column.

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided.

        Parameters
        ----------
        aggregation_func : callable, optional, default ``numpy.nanmean``
            Function that aggregates an array to a number.
            Signature (y: array) -> aggregated value: float.
        aggregation_func_name : `str` or None, optional, default "mean"
            Name of grouping function, used to report results.
            If None, defaults to "aggregation".
        groupby_time_feature : `str` or None, optional
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, optional
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, optional
            If provided, groups by this column value. Should be same length as the DataFrame.

        Returns
        -------
        grouped_df : `pandas.DataFrame` with two columns:

            (1) grouping_func_name:
                evaluation metric for aggregation of timeseries.
            (2) group name:
                group name depends on the grouping method:
                ``groupby_time_feature`` for ``groupby_time_feature``
                ``cst.TIME_COL`` for ``groupby_sliding_window_size``
                ``groupby_custom_column.name`` for ``groupby_custom_column``.
        """
        df = self.df.copy()
        if aggregation_func_name:
            grouping_func_name = f"{aggregation_func_name} of {VALUE_COL}"
        else:
            grouping_func_name = f"aggregation of {VALUE_COL}"

        def grouping_func(grp):
            return aggregation_func(grp[VALUE_COL])

        result = add_groupby_column(
            df=df,
            time_col=TIME_COL,
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column)

        grouped_df = grouping_evaluation(
            df=result["df"],
            groupby_col=result["groupby_col"],
            grouping_func=grouping_func,
            grouping_func_name=grouping_func_name)
        return grouped_df

    def plot_grouping_evaluation(
            self,
            aggregation_func=np.nanmean,
            aggregation_func_name="mean",
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None,
            xlabel=None,
            ylabel=None,
            title=None):
        """Computes aggregated timeseries by group and plots the result.
        Can be used to plot aggregated timeseries by a time feature, over time,
        or by a user-provided column.

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided.

        Parameters
        ----------
        aggregation_func : callable, optional, default ``numpy.nanmean``
            Function that aggregates an array to a number.
            Signature (y: array) -> aggregated value: float.
        aggregation_func_name : `str` or None, optional, default "mean"
            Name of grouping function, used to report results.
            If None, defaults to "aggregation".
        groupby_time_feature : `str` or None, optional
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, optional
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, optional
            If provided, groups by this column value. Should be same length as the DataFrame.
        xlabel : `str`, optional, default None
            X-axis label of the plot.
        ylabel : `str`, optional, default None
            Y-axis label of the plot.
        title : `str` or None, optional
            Plot title. If None, default is based on axis labels.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            plotly graph object showing aggregated timeseries by group.
            x-axis label depends on the grouping method:
            ``groupby_time_feature`` for ``groupby_time_feature``
            ``TIME_COL`` for ``groupby_sliding_window_size``
            ``groupby_custom_column.name`` for ``groupby_custom_column``.
        """
        grouped_df = self.get_grouping_evaluation(
            aggregation_func=aggregation_func,
            aggregation_func_name=aggregation_func_name,
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column)

        xcol, ycol = grouped_df.columns
        fig = plot_univariate(
            df=grouped_df,
            x_col=xcol,
            y_col=ycol,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title)

        return fig

    def get_quantiles_and_overlays(
            self,
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None,
            show_mean=False,
            show_quantiles=False,
            show_overlays=False,
            overlay_label_time_feature=None,
            overlay_label_sliding_window_size=None,
            overlay_label_custom_column=None,
            center_values=False,
            value_col=VALUE_COL,
            mean_col_name="mean",
            quantile_col_prefix="Q",
            **overlay_pivot_table_kwargs):
        """Computes mean, quantiles, and overlays by the requested grouping dimension.

        Overlays are best explained in the plotting context. The grouping dimension goes on
        the x-axis, and one line is shown for each level of the overlay dimension. This
        function returns a column for each line to plot (e.g. mean, each quantile,
        each overlay value).

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided as the grouping dimension.

        If ``show_overlays`` is True, exactly one of: ``overlay_label_time_feature``,
        ``overlay_label_sliding_window_size``, ``overlay_label_custom_column`` can be
        provided to specify the ``label_col`` (overlay dimension). Internally, the
        function calls `pandas.DataFrame.pivot_table` with ``index=groupby_col``,
        ``columns=label_col``, ``values=value_col`` to get the overlay values for plotting.
        You can pass additional parameters to `pandas.DataFrame.pivot_table` via
        ``overlay_pivot_table_kwargs``, e.g. to change the aggregation method. If an explicit
        label is not provided, the records are labeled by their position within the group.

        For example, to show yearly seasonality mean, quantiles, and overlay plots for
        each individual year, use::

            self.get_quantiles_and_overlays(
                groupby_time_feature="doy",         # Rows: a row for each day of year (1, 2, ..., 366)
                show_mean=True,                     # mean value on that day
                show_quantiles=[0.1, 0.9],          # quantiles of the observed distribution on that day
                show_overlays=True,                 # Include overlays defined by ``overlay_label_time_feature``
                overlay_label_time_feature="year")  # One column for each observed "year" (2016, 2017, 2018, ...)

        To show weekly seasonality over time, use::

            self.get_quantiles_and_overlays(
                groupby_time_feature="dow",            # Rows: a row for each day of week (1, 2, ..., 7)
                show_mean=True,                        # mean value on that day
                show_quantiles=[0.1, 0.5, 0.9],        # quantiles of the observed distribution on that day
                show_overlays=True,                    # Include overlays defined by ``overlay_label_time_feature``
                overlay_label_sliding_window_size=90,  # One column for each 90 period sliding window in the dataset,
                aggfunc="median")                      # overlay value is the median value for the dow over the period (default="mean").


        It may be difficult to assess the weekly seasonality from the previous result,
        because overlays shift up/down over time due to trend/yearly seasonality.
        Use ``center_values=True`` to adjust each overlay so its average value is centered at 0.
        Mean and quantiles are shifted by a single constant to center the mean at 0, while
        preserving their relative values::

            self.get_quantiles_and_overlays(
                groupby_time_feature="dow",
                show_mean=True,
                show_quantiles=[0.1, 0.5, 0.9],
                show_overlays=True,
                overlay_label_sliding_window_size=90,
                aggfunc="median",
                center_values=True)  # Centers the output

        Centering reduces the variability in the overlays to make it easier to isolate
        the effect by the groupby column. As a result, centered overlays have smaller
        variability than that reported by the quantiles, which operate on the original,
        uncentered data points. Similarly, if overlays are aggregates of individual values
        (i.e. ``aggfunc`` is needed in the call to `pandas.DataFrame.pivot_table`),
        the quantiles of overlays will be less extreme than those of the original data.

            - To assess variability conditioned on the groupby value, check the quantiles.
            - To assess variability conditioned on both the groupby and overlay value,
              after any necessary aggregation, check the variability of the overlay values.
              Compute quantiles of overlays from the return value if desired.

        Parameters
        ----------
        groupby_time_feature : `str` or None, default None
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, default None
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, default None
            If provided, groups by this column value. Should be same length as the DataFrame.
        show_mean : `bool`, default False
            Whether to return the mean value by the groupby column.
        show_quantiles : `bool` or `list` [`float`] or `numpy.array`, default False
            Whether to return the quantiles of the value by the groupby column.
            If False, does not return quantiles. If True, returns default
            quantiles (0.1 and 0.9). If array-like, a list of quantiles
            to compute (e.g. (0.1, 0.25, 0.75, 0.9)).
        show_overlays : `bool` or `int` or array-like [`int` or `str`] or None, default False
            Whether to return overlays of the value by the groupby column.

            If False, no overlays are shown.

            If True and ``label_col`` is defined, calls `pandas.DataFrame.pivot_table` with
            ``index=groupby_col``, ``columns=label_col``, ``values=value_col``.
            ``label_col`` is defined by one of ``overlay_label_time_feature``,
            ``overlay_label_sliding_window_size``, or ``overlay_label_custom_column``.
            Returns one column for each value of the ``label_col``.

            If True and the ``label_col`` is not defined, returns the raw values within
            each group. Values across groups are put into columns by their position in
            the group (1st element in group, 2nd, 3rd, etc.). Positional order in a group
            is not guaranteed to correspond to anything meaningful, so the items within a
            column may not have anything in common. It is better to specify one of ``overlay_*``
            to explicitly define the overlay labels.

            If an integer, the number of overlays to randomly sample. The same as True,
            then randomly samples up to `int` columns. This is useful if there are too many values.

            If a list [int], a list of column indices (int type). The same as True,
            then selects the specified columns by index.

            If a list [str], a list of column names. Column names are matched by their
            string representation to the names in this list. The same as True,
            then selects the specified columns by name.
        overlay_label_time_feature : `str` or None, default None
            If ``show_overlays`` is True, can be used to define ``label_col``,
            i.e. which dimension to show separately as overlays.

            If provided, uses a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        overlay_label_sliding_window_size : `int` or None, default None
            If ``show_overlays`` is True, can be used to define ``label_col``,
            i.e. which dimension to show separately as overlays.

            If provided, uses a column that sequentially partitions data into groups
            of size ``groupby_sliding_window_size``.
        overlay_label_custom_column : `pandas.Series` or None, default None
            If ``show_overlays`` is True, can be used to define ``label_col``,
            i.e. which dimension to show separately as overlays.

            If provided, uses this column value. Should be same length as the DataFrame.

        value_col : `str`, default VALUE_COL
            The column name for the value column. By default,
            shows the univariate time series value, but it can be any
            other column in ``self.df``.

        mean_col_name : `str`, default "mean"
            The name to use for the mean column in the output.
            Applies if ``show_mean=True``.

        quantile_col_prefix : `str`, default "Q"
            The prefix to use for quantile column names in the output.
            Columns are named with this prefix followed by the quantile,
            rounded to 2 decimal places.

        center_values : `bool`, default False
            Whether to center the return values.
            If True, shifts each overlay so its average value is centered at 0.
            Shifts mean and quantiles by a constant to center the mean at 0, while
            preserving their relative values.

            If False, values are not centered.

        overlay_pivot_table_kwargs : additional parameters
            Additional keyword parameters to pass to `pandas.DataFrame.pivot_table`,
            used in generating the overlays. See above description for details.

        Returns
        -------
        grouped_df : `pandas.DataFrame`
            Dataframe with mean, quantiles, and overlays by the grouping column. Overlays
            are defined by the grouping column and overlay dimension.

            ColumnIndex is a multiindex with first level as the "category", a subset of
            [MEAN_COL_GROUP, QUANTILE_COL_GROUP, OVERLAY_COL_GROUP] depending on what is requests.

                - grouped_df[MEAN_COL_GROUP] = df with single column, named ``mean_col_name``.

                - grouped_df[QUANTILE_COL_GROUP] = df with a column for each quantile, named
                  f"{quantile_col_prefix}{round(str(q))}", where ``q`` is the quantile.

                - grouped_df[OVERLAY_COL_GROUP] = df with one column per overlay value, named
                  by the overlay value.

            For example, it might look like::

                category    mean    quantile        overlay
                name        mean    Q0.1    Q0.9    2007    2008    2009
                doy
                1	        8.42	7.72    9.08	8.29	7.75	8.33
                2	        8.82	8.20    9.56	8.43	8.80	8.53
                3	        8.95	8.25    9.88	8.26	9.12	8.70
                4	        9.07	8.60    9.49	8.10	9.99	8.73
                5	        8.73	8.29    9.24	7.95	9.26	8.37
                ...         ...     ...     ...     ...     ...     ...

        """
        # Default quantiles to show if `show_quantiles` is boolean
        if isinstance(show_quantiles, bool):
            if show_quantiles:
                show_quantiles = [0.1, 0.9]
            else:
                show_quantiles = None

        # Adds grouping dimension
        result = add_groupby_column(
            df=self.df,
            time_col=TIME_COL,  # Already standardized
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column)
        df = result["df"]
        groupby_col = result["groupby_col"]
        grouped_df = None

        # Whether an overlay label is provided
        add_overlay_label = (overlay_label_time_feature is not None) or \
            (overlay_label_sliding_window_size is not None) or \
            (overlay_label_custom_column is not None)
        overlay_df = None

        # Defines an aggregation function to compute mean, quantiles, and overlays
        agg_kwargs = {}
        if show_mean:
            agg_kwargs.update({mean_col_name: pd.NamedAgg(column=value_col, aggfunc=np.nanmean)})
        if show_quantiles is not None:
            # Returns the quantiles of the group's `value_col` as a list
            agg_kwargs.update({quantile_col_prefix: pd.NamedAgg(
                column=value_col,
                aggfunc=lambda grp_values: partial(np.nanquantile, q=show_quantiles)(grp_values).tolist())})
        if show_overlays is not False:
            if add_overlay_label:
                # Uses DataFrame pivot_table to get overlay labels as columns, `groupby_col` as index
                label_result = add_groupby_column(
                    df=df,
                    time_col=TIME_COL,
                    groupby_time_feature=overlay_label_time_feature,
                    groupby_sliding_window_size=overlay_label_sliding_window_size,
                    groupby_custom_column=overlay_label_custom_column)
                label_col = label_result["groupby_col"]
                overlay_df = label_result["df"].pivot_table(
                    index=groupby_col,
                    columns=label_col,
                    values=value_col,
                    **overlay_pivot_table_kwargs)
            else:
                # Uses aggregation to get overlays.
                # Takes original values within each group.
                # Values across groups are put into columns by their position
                # within the group (1st element in group, 2nd, 3rd, etc.)
                agg_kwargs.update({"overlay": pd.NamedAgg(column=value_col, aggfunc=tuple)})

        # Names the quantile columns
        # Keeps to 2 decimal places to handle numerical imprecision.
        list_names_dict = {quantile_col_prefix: [
            f"{quantile_col_prefix}{str(round(x, 2))}" for x in show_quantiles]}\
            if show_quantiles is not None else {}
        if agg_kwargs:
            grouped_df = flexible_grouping_evaluation(
                result["df"],
                map_func_dict=None,
                groupby_col=result["groupby_col"],
                agg_kwargs=agg_kwargs,
                extend_col_names=False,
                unpack_list=True,
                list_names_dict=list_names_dict)
        # Adds overlays if requested and not already computed during aggregation
        if overlay_df is not None:
            overlay_df.columns = map(str, overlay_df.columns)

        # Either overlay_df or grouped_df is populated
        if grouped_df is None and overlay_df is None:
            raise ValueError("Must enable at least one of: show_mean, show_quantiles, show_overlays.")
        grouped_df = pd.concat([grouped_df, overlay_df], axis=1)

        # Creates MultiIndex for column names to categorize the column names by their type
        mean_cols = [mean_col_name] if show_mean else []
        quantile_cols = list_names_dict.get(quantile_col_prefix, [])
        overlay_cols = [col for col in list(grouped_df.columns) if col not in mean_cols + quantile_cols]

        if isinstance(show_overlays, int) and not isinstance(show_overlays, bool):
            # Samples from `overlay_cols`
            which_overlays = sorted(np.random.choice(
                range(len(overlay_cols)),
                size=min(show_overlays, len(overlay_cols)),
                replace=False))
            overlay_cols = list(np.array(overlay_cols)[which_overlays])
        elif isinstance(show_overlays, (list, tuple, np.ndarray)):
            # Selects from `overlay_cols`
            all_integers = np.issubdtype(np.array(show_overlays).dtype, np.integer)
            if all_integers:
                overlay_cols = [col for i, col in enumerate(overlay_cols) if i in show_overlays]
            else:
                overlay_cols = [col for col in overlay_cols if str(col) in show_overlays]
        cols = mean_cols + quantile_cols + overlay_cols  # Reorders columns by group
        grouped_df = grouped_df[cols]
        categories = list(np.repeat(
            [MEAN_COL_GROUP, QUANTILE_COL_GROUP, OVERLAY_COL_GROUP],  # Labels columns by category
            [len(mean_cols), len(quantile_cols), len(overlay_cols)]))
        cateory_col_index = pd.MultiIndex.from_arrays([categories, cols], names=["category", "name"])
        grouped_df.columns = cateory_col_index

        if center_values:
            # Each overlay is independently shifted to have mean 0.
            if OVERLAY_COL_GROUP in grouped_df:
                grouped_df[OVERLAY_COL_GROUP] -= grouped_df[OVERLAY_COL_GROUP].mean()
            # Mean and quantiles are shifted by the same constant, so the mean column is centered at 0.
            if MEAN_COL_GROUP in grouped_df:
                mean_shift = grouped_df[MEAN_COL_GROUP].mean()[0]
                grouped_df[MEAN_COL_GROUP] -= mean_shift
            else:
                mean_shift = self.df[value_col].mean()
            if QUANTILE_COL_GROUP in grouped_df:
                grouped_df[QUANTILE_COL_GROUP] -= mean_shift

        return grouped_df

    def plot_quantiles_and_overlays(
            self,
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None,
            show_mean=False,
            show_quantiles=False,
            show_overlays=False,
            overlay_label_time_feature=None,
            overlay_label_sliding_window_size=None,
            overlay_label_custom_column=None,
            center_values=False,
            value_col=VALUE_COL,
            mean_col_name="mean",
            quantile_col_prefix="Q",
            mean_style=None,
            quantile_style=None,
            overlay_style=None,
            xlabel=None,
            ylabel=None,
            title=None,
            showlegend=True,
            **overlay_pivot_table_kwargs):
        """Plots mean, quantiles, and overlays by the requested grouping dimension.

        The grouping dimension goes on the x-axis, and one line is shown for the mean,
        each quantile, and each level of the overlay dimension, as requested. By default,
        shading is applied between the quantiles.

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided as the grouping dimension.

        If ``show_overlays`` is True, exactly one of: ``overlay_label_time_feature``,
        ``overlay_label_sliding_window_size``, ``overlay_label_custom_column`` can be
        provided to specify the ``label_col`` (overlay dimension). Internally, the
        function calls `pandas.DataFrame.pivot_table` with ``index=groupby_col``,
        ``columns=label_col``, ``values=value_col`` to get the overlay values for plotting.
        You can pass additional parameters to `pandas.DataFrame.pivot_table` via
        ``overlay_pivot_table_kwargs``, e.g. to change the aggregation method. If an explicit
        label is not provided, the records are labeled by their position within the group.

        For example, to show yearly seasonality mean, quantiles, and overlay plots for
        each individual year, use::

            self.plot_quantiles_and_overlays(
                groupby_time_feature="doy",         # Rows: a row for each day of year (1, 2, ..., 366)
                show_mean=True,                     # mean value on that day
                show_quantiles=[0.1, 0.9],          # quantiles of the observed distribution on that day
                show_overlays=True,                 # Include overlays defined by ``overlay_label_time_feature``
                overlay_label_time_feature="year")  # One column for each observed "year" (2016, 2017, 2018, ...)

        To show weekly seasonality over time, use::

            self.plot_quantiles_and_overlays(
                groupby_time_feature="dow",            # Rows: a row for each day of week (1, 2, ..., 7)
                show_mean=True,                        # mean value on that day
                show_quantiles=[0.1, 0.5, 0.9],        # quantiles of the observed distribution on that day
                show_overlays=True,                    # Include overlays defined by ``overlay_label_time_feature``
                overlay_label_sliding_window_size=90,  # One column for each 90 period sliding window in the dataset,
                aggfunc="median")                      # overlay value is the median value for the dow over the period (default="mean").

        It may be difficult to assess the weekly seasonality from the previous result,
        because overlays shift up/down over time due to trend/yearly seasonality.
        Use ``center_values=True`` to adjust each overlay so its average value is centered at 0.
        Mean and quantiles are shifted by a single constant to center the mean at 0, while
        preserving their relative values::

            self.plot_quantiles_and_overlays(
                groupby_time_feature="dow",
                show_mean=True,
                show_quantiles=[0.1, 0.5, 0.9],
                show_overlays=True,
                overlay_label_sliding_window_size=90,
                aggfunc="median",
                center_values=True)  # Centers the output

        Centering reduces the variability in the overlays to make it easier to isolate
        the effect by the groupby column. As a result, centered overlays have smaller
        variability than that reported by the quantiles, which operate on the original,
        uncentered data points. Similarly, if overlays are aggregates of individual values
        (i.e. ``aggfunc`` is needed in the call to `pandas.DataFrame.pivot_table`),
        the quantiles of overlays will be less extreme than those of the original data.

            - To assess variability conditioned on the groupby value, check the quantiles.
            - To assess variability conditioned on both the groupby and overlay value,
              after any necessary aggregation, check the variability of the overlay values.
              Compute quantiles of overlays from the return value if desired.

        Parameters
        ----------
        groupby_time_feature : `str` or None, default None
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, default None
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, default None
            If provided, groups by this column value. Should be same length as the DataFrame.
        show_mean : `bool`, default False
            Whether to return the mean value by the groupby column.
        show_quantiles : `bool` or `list` [`float`] or `numpy.array`, default False
            Whether to return the quantiles of the value by the groupby column.
            If False, does not return quantiles. If True, returns default
            quantiles (0.1 and 0.9). If array-like, a list of quantiles
            to compute (e.g. (0.1, 0.25, 0.75, 0.9)).
        show_overlays : `bool` or `int` or array-like [`int` or `str`], default False
            Whether to return overlays of the value by the groupby column.

            If False, no overlays are shown.

            If True and ``label_col`` is defined, calls `pandas.DataFrame.pivot_table` with
            ``index=groupby_col``, ``columns=label_col``, ``values=value_col``.
            ``label_col`` is defined by one of ``overlay_label_time_feature``,
            ``overlay_label_sliding_window_size``, or ``overlay_label_custom_column``.
            Returns one column for each value of the ``label_col``.

            If True and the ``label_col`` is not defined, returns the raw values within
            each group. Values across groups are put into columns by their position in
            the group (1st element in group, 2nd, 3rd, etc.). Positional order in a group
            is not guaranteed to correspond to anything meaningful, so the items within a
            column may not have anything in common. It is better to specify one of ``overlay_*``
            to explicitly define the overlay labels.

            If an integer, the number of overlays to randomly sample. The same as True,
            then randomly samples up to `int` columns. This is useful if there are too many values.

            If a list [int], a list of column indices (int type). The same as True,
            then selects the specified columns by index.

            If a list [str], a list of column names. Column names are matched by their
            string representation to the names in this list. The same as True,
            then selects the specified columns by name.

        overlay_label_time_feature : `str` or None, default None
            If ``show_overlays`` is True, can be used to define ``label_col``,
            i.e. which dimension to show separately as overlays.

            If provided, uses a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        overlay_label_sliding_window_size : `int` or None, default None
            If ``show_overlays`` is True, can be used to define ``label_col``,
            i.e. which dimension to show separately as overlays.

            If provided, uses a column that sequentially partitions data into groups
            of size ``groupby_sliding_window_size``.
        overlay_label_custom_column : `pandas.Series` or None, default None
            If ``show_overlays`` is True, can be used to define ``label_col``,
            i.e. which dimension to show separately as overlays.

            If provided, uses this column value. Should be same length as the DataFrame.

        value_col : `str`, default VALUE_COL
            The column name for the value column. By default,
            shows the univariate time series value, but it can be any
            other column in ``self.df``.

        mean_col_name : `str`, default "mean"
            The name to use for the mean column in the output.
            Applies if ``show_mean=True``.

        quantile_col_prefix : `str`, default "Q"
            The prefix to use for quantile column names in the output.
            Columns are named with this prefix followed by the quantile,
            rounded to 2 decimal places.

        center_values : `bool`, default False
            Whether to center the return values.
            If True, shifts each overlay so its average value is centered at 0.
            Shifts mean and quantiles by a constant to center the mean at 0, while
            preserving their relative values.

            If False, values are not centered.

        mean_style: `dict` or None, default None
            How to style the mean line, passed as keyword arguments to
            `plotly.graph_objs.Scatter`. If None, the default is::

                mean_style = {
                    "line": dict(
                        width=2,
                        color="#595959"),  # gray
                    "legendgroup": MEAN_COL_GROUP}

        quantile_style: `dict` or None, default None
            How to style the quantile lines, passed as keyword arguments to
            `plotly.graph_objs.Scatter`. If None, the default is::

                quantile_style = {
                    "line": dict(
                        width=2,
                        color="#1F9AFF",  # blue
                        dash="solid"),
                    "legendgroup": QUANTILE_COL_GROUP,  # show/hide them together
                    "fill": "tonexty"}

            Note that fill style is removed from to the first quantile line, to
            fill only between items in the same category.

        overlay_style: `dict` or None, default None
            How to style the overlay lines, passed as keyword arguments to
            `plotly.graph_objs.Scatter`. If None, the default is::

                overlay_style = {
                    "opacity": 0.5,  # makes it easier to see density
                    "line": dict(
                        width=1,
                        color="#B3B3B3",  # light gray
                        dash="solid"),
                    "legendgroup": OVERLAY_COL_GROUP}

        xlabel : `str`, optional, default None
            X-axis label of the plot.
        ylabel : `str`, optional, default None
            Y-axis label of the plot. If None, uses ``value_col``.
        title : `str` or None, default None
            Plot title. If None, default is based on axis labels.
        showlegend : `bool`, default True
            Whether to show the legend.
        overlay_pivot_table_kwargs : additional parameters
            Additional keyword parameters to pass to `pandas.DataFrame.pivot_table`,
            used in generating the overlays.
            See `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries.get_quantiles_and_overlays`
            description for details.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            plotly graph object showing the mean, quantiles, and overlays.

        See Also
        --------
        `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries.get_quantiles_and_overlays`
            To get the mean, quantiles, and overlays as a `pandas.DataFrame` without plotting.
        """

        if ylabel is None:
            ylabel = value_col

        grouped_df = self.get_quantiles_and_overlays(
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column,
            show_mean=show_mean,
            show_quantiles=show_quantiles,
            show_overlays=show_overlays,
            overlay_label_time_feature=overlay_label_time_feature,
            overlay_label_sliding_window_size=overlay_label_sliding_window_size,
            overlay_label_custom_column=overlay_label_custom_column,
            center_values=center_values,
            value_col=value_col,
            mean_col_name=mean_col_name,
            quantile_col_prefix=quantile_col_prefix,
            **overlay_pivot_table_kwargs)

        if mean_style is None:
            mean_style = {
                "line": dict(
                    width=2,
                    color="#595959"),  # gray
                "legendgroup": MEAN_COL_GROUP}
        if quantile_style is None:
            quantile_style = {
                "line": dict(
                    width=2,
                    color="#1F9AFF",  # blue
                    dash="solid"),
                "legendgroup": QUANTILE_COL_GROUP,  # show/hide them together
                "fill": "tonexty"}
        if overlay_style is None:
            overlay_style = {
                "opacity": 0.5,  # makes it easier to see density
                "line": dict(
                    width=1,
                    color="#B3B3B3",  # light gray
                    dash="solid"),
                "legendgroup": OVERLAY_COL_GROUP}
        style_dict = {
            MEAN_COL_GROUP: mean_style,
            QUANTILE_COL_GROUP: quantile_style,
            OVERLAY_COL_GROUP: overlay_style}

        y_col_style_dict = {}
        # All categories in grouped_df. Reverses the order so the first category is plotted last (on top).
        categories = grouped_df.columns.get_level_values(0).unique()[::-1]
        for category in categories:
            style = style_dict.get(category, {})
            if "fill" in style:
                # If fill is part of the style, plotly fills the area between this line and
                # the previous line added to the plot.
                # Since we only want to fill between lines in the same category (e.g. between quantiles),
                # we remove the "fill" from the first line within each category. Otherwise the first
                # line in this category would fill to the last line in the previous category.
                category_style_dict = {grouped_df[category].columns[0]: {k: v for k, v in style.items() if k != "fill"}}
                category_style_dict.update({col: style for col in grouped_df[category].columns[1:]})
            else:
                category_style_dict = {col: style for col in grouped_df[category].columns}
            y_col_style_dict.update(category_style_dict)

        grouped_df.columns = list(grouped_df.columns.get_level_values(1))  # MultiIndex is not needed for plotting
        x_col = grouped_df.index.name
        grouped_df.reset_index(inplace=True)
        fig = plot_multivariate(
            grouped_df,
            x_col=x_col,
            y_col_style_dict=y_col_style_dict,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            showlegend=showlegend)
        return fig



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Kaixu Yang
"""Utilities for changepoint detection via adaptive lasso."""

import warnings
from datetime import timedelta

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from pandas.plotting import register_matplotlib_converters
from pandas.tseries.frequencies import to_offset
from sklearn.linear_model import Lasso
from sklearn.linear_model import LassoCV
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import RidgeCV

from greykite.common.constants import CHANGEPOINT_COL_PREFIX
from greykite.common.features.timeseries_features import add_time_features_df
from greykite.common.features.timeseries_features import build_time_features_df
from greykite.common.features.timeseries_features import fourier_series_multi_fcn
from greykite.common.features.timeseries_features import get_changepoint_dates_from_changepoints_dict
from greykite.common.features.timeseries_features import get_changepoint_features
from greykite.common.features.timeseries_features import get_changepoint_features_and_values_from_config
from greykite.common.features.timeseries_features import get_default_origin_for_time_vars
from greykite.common.python_utils import get_pattern_cols


np.seterr(divide="ignore")  # np.where evaluates values before selecting by conditions, set this to suppress divide_by_zero error


register_matplotlib_converters()


def check_freq_unit_at_most_day(freq, name):
    """Checks if the ``freq`` parameter passed to a function has unit at most "D"

    Parameters
    ----------
    freq : `DateOffset`, `Timedelta` or `str`
        The parameter passed as ``freq``.
    name : `str`
        The name of the parameter ``freq``.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        Because the input frequency has unit greater than "D".
    """
    if (isinstance(freq, str) and
            any(char in freq for char in ["W", "M", "Y"])):
        raise ValueError(f"In {name}, the maximal unit is 'D', "
                         "i.e., you may use units no more than 'D' such as"
                         "'10D', '5H', '100T', '200S'. The reason is that 'W', 'M' "
                         "or higher has either cycles or indefinite number of days, "
                         "thus is not parsable by pandas as timedelta.")


def build_trend_feature_df_with_changes(
        df,
        time_col,
        origin_for_time_vars=None,
        changepoints_dict="auto"):
    """A function to generate trend features from a given time series df.

    The trend features include columns of the format max(0, x-c_i), where c_i is the i-th change point value.
    If changepoints_dict has "method": "uniform", then n_changepoints change points plus the normal growth term are
    generated uniformly over the whole time period. This is recommended.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data df.
    time_col : `str'
        The column name of time column in ``df``, entries can be parsed with pd.to_datetime.
    origin_for_time_vars : `float` or `None`, default `None`
        Original continuous time value, if not provided, will be parsed from data.
    changepoints_dict: `str`: "auto" or `dict`, default "auto"
        Change point dictionary, compatible with
        `~greykite.common.features.timeseries_features.get_changepoint_features_and_values_from_config`
        If not provided, default is 100 change points evenly
        distributed over the whole time period.

    Returns
    -------
    df :  `pandas.DataFrame`
        Change point feature df with n_changepoints + 1 columns generated.
    """
    df = df.copy()
    # Gets changepoints features from config. Evenly distributed with n_changepoints specified is recommended.
    if origin_for_time_vars is None:
        origin_for_time_vars = get_default_origin_for_time_vars(df, time_col)
    if changepoints_dict == "auto":
        changepoints_dict = {
            "method": "uniform",
            "n_changepoints": 100
        }
    changepoints = get_changepoint_features_and_values_from_config(
        df=df,
        time_col=time_col,
        changepoints_dict=changepoints_dict,
        origin_for_time_vars=origin_for_time_vars)
    if changepoints["changepoint_values"] is None:
        # allows n_changepoints = 0
        changepoint_values = np.array([0])
    else:
        changepoint_values = np.concatenate([[0], changepoints["changepoint_values"]])
    growth_func = changepoints["growth_func"]
    features_df = add_time_features_df(
        df=df,
        time_col=time_col,
        conti_year_origin=origin_for_time_vars)
    changepoint_dates = get_changepoint_dates_from_changepoints_dict(
        changepoints_dict=changepoints_dict,
        df=df,
        time_col=time_col)
    changepoint_dates = [pd.to_datetime(df[time_col].iloc[0])] + changepoint_dates
    changepoint_features_df = get_changepoint_features(
        features_df,
        changepoint_values,
        continuous_time_col="ct1",
        growth_func=growth_func,
        changepoint_dates=changepoint_dates)
    changepoint_features_df.index = pd.to_datetime(df[time_col])
    return changepoint_features_df


def build_seasonality_feature_df_with_changes(
        df,
        time_col,
        origin_for_time_vars=None,
        changepoints_dict=None,
        fs_components_df=pd.DataFrame({
            "name": ["tod", "tow", "toy"],
            "period": [24.0, 7.0, 1.0],
            "order": [3, 3, 5],
            "seas_names": ["daily", "weekly", "yearly"]})):
    """A function to generate yearly seasonality features from a given time series df.

    The seasonality features include n_changepoints * 2 columns of the format
    1{x > c_i} * sin(2 * pi / period * order * x) and 1{x > c_i} * cos(2 * pi / period * order * x),
    where c_i is the i-th change point value.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data df.
    time_col : `str'
        The column name of time column in ``df``, entries can be parsed with pd.to_datetime.
    origin_for_time_vars : `float` or `None`, default `None`
        Original continuous time value, if not provided, will be parsed from data.
    changepoints_dict: `dict` or `None`, default `None`
        Change point dictionary, compatible with
        `~greykite.common.features.timeseries_features.get_changepoint_features_and_values_from_config`
        If not provided, default is no change points.
    fs_components_df: `pandas.DataFrame`
        fs_components config df, compatible with
        `~greykite.common.features.timeseries_features.fourier_series_multi_fcn`

    Returns
    -------
    df : `pandas.DataFrame`
        Seasonality feature df with [sum_{component_i} (2 * order_of_component_i)] * (num_changepoints + 1)
        columns.
        Each sum_{component_i} (2 * order_of_component_i) columns form a block. The first block contains the
        original seasonality features, which accounts for the overall seasonality magnitudes. Each of the
        rest num_changepoints block is a copy of the first block, with rows whose timestamps are before the
        corresponding changepoint replaced by zeros. Such blocks only have effect on the seasonality
        magnitudes after the corresponding changepoints.
    """
    df = df.copy()
    # Gets changepoints features from config. Evenly distributed with n_changepoints specified is recommended.
    if origin_for_time_vars is None:
        origin_for_time_vars = get_default_origin_for_time_vars(df, time_col)
    features_df = add_time_features_df(
        df=df,
        time_col=time_col,
        conti_year_origin=origin_for_time_vars)
    fs_func = None
    fs_cols = []
    if fs_components_df is not None:
        fs_components_df = fs_components_df[fs_components_df["order"] != 0]
        fs_components_df = fs_components_df.reset_index()
        if fs_components_df.shape[0] > 0:
            fs_func = fourier_series_multi_fcn(
                col_names=fs_components_df["name"],
                periods=fs_components_df.get("period"),
                orders=fs_components_df.get("order"),
                seas_names=fs_components_df.get("seas_names")
            )
            time_features_example_df = build_time_features_df(
                df[time_col][:2],  # only needs the first two rows to get the column info
                conti_year_origin=origin_for_time_vars)
            fs = fs_func(time_features_example_df)
            fs_cols = fs["cols"]
    fs_features = fs_func(features_df)
    fs_df = fs_features["df"]
    fs_df.index = pd.to_datetime(df[time_col])
    # Augments regular Fourier columns to truncated Fourier columns.
    if changepoints_dict is not None:
        changepoint_dates = get_changepoint_dates_from_changepoints_dict(
            changepoints_dict=changepoints_dict,
            df=df,
            time_col=time_col
        )
        # The following lines truncates the fourier series at each change point
        # For each change point, the values of the fourier series before the change point are
        # set to zero. The column name is simply appending `_%Y_%m_%d_%H` after the original column names.
        col_names = fs_cols + [f"{col}{date.strftime('_%Y_%m_%d_%H')}" for date in changepoint_dates for col in fs_cols]
        fs_truncated_df = pd.concat([fs_df] * (len(changepoint_dates) + 1), axis=1)
        fs_truncated_df.columns = col_names
        for i, date in enumerate(changepoint_dates):
            cols = fs_truncated_df.columns[fs_df.shape[1] * (i + 1):]
            fs_truncated_df.loc[(features_df["datetime"] < date).values, cols] = 0
        fs_df = fs_truncated_df
    return fs_df


def build_seasonality_feature_df_from_detection_result(
        df,
        time_col,
        seasonality_changepoints,
        seasonality_components_df,
        include_original_block=True,
        include_components=None):
    """Builds seasonality feature df from detected seasonality change points.

    The function is an extension of ``build_seasonality_feature_df_with_changes``. It can generate
    a seasonality feature df with different change points for different components. These changepoints
    are passed through a dictionary, which can be the output of
    `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`

    Parameters
    __________
    df : `pandas.DataFrame`
        The dataframe used to build seasonality feature df.
    time_col : `str`
        The name of the time column in ``df``.
    seasonality_changepoints : `dict`
        The seasonality change point dictionary. The keys are seasonality components, and the values
        are the corresponding change points given in lists.
        For example

            "weekly": [Timestamp('2020-01-01 00:00:00'), Timestamp('2021-04-05 00:00:00')]
            "yearly": [Timestamp('2020-08-06 00:00:00')]

    seasonality_components_df : `pandas.DataFrame`
        The seasonality components dataframe that is compatible with
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`
        The values in "seas_names" must equal to the keys in ``seasonality_changepoints``.
    include_original_block : `bool`, default True
        Whether to include the original untruncated block of sin or cos columns for each component. If set to
        False, the original seasonality block for each component will be dropped.
    include_components : `list` [`str`] or None, default None
        The components to be included from the result. If None, all components will be included.

    Returns
    -------
    seasonality_feature_df : `pandas.DataFrame`
        The seasonality feature dataframe similar to the output of `build_seasonality_feature_df_with_changes``
        but possibly with different changepoints on different components.
    """
    seasonality_df = pd.DataFrame()
    for component in seasonality_changepoints.keys():
        if include_components is None or component in include_components:
            seasonality_df_cp = build_seasonality_feature_df_with_changes(
                df=df,
                time_col=time_col,
                changepoints_dict={
                    "method": "custom",
                    "dates": seasonality_changepoints[component]
                },
                fs_components_df=seasonality_components_df[seasonality_components_df["seas_names"] == component]
            )
            if not include_original_block:
                original_block_size = int(2 * seasonality_components_df.loc[
                    seasonality_components_df["seas_names"] == component, "order"].values[0])
                seasonality_df_cp = seasonality_df_cp.iloc[:, original_block_size:]
            seasonality_df = pd.concat([seasonality_df, seasonality_df_cp], axis=1)
    extra_components = [component for component in seasonality_changepoints if component not in include_components] if include_components is not None else []
    if extra_components:
        warnings.warn(f"The following seasonality components have detected seasonality changepoints"
                      f" but these changepoints are not included in the model,"
                      f" because the seasonality component is not included in the model. {extra_components}")
    return seasonality_df


def compute_fitted_components(
        x,
        coef,
        regex,
        include_intercept,
        intercept=None):
    """Computes the fitted values with selected regressors indicated by ``regex``

    Parameters
    ----------
    x : `pandas.DataFrame`
        The design matrix df with conventional column names.
    coef : `numpy.array`
        Estimated coefficients.
    regex : regular expression
        Pattern of the names of the columns to be used.
    include_intercept : bool
        Whether to include intercept.
    intercept : `float`
        The estimated intercept, must be provided if ``include_intercept`` == True.

    Returns
    -------
    result: `pandas.Series`
        The estimated component from selected regressors.
    """
    if include_intercept and intercept is None:
        raise ValueError("``intercept`` must be provided when ``include_intercept`` is True.")
    columns = list(x.columns)
    selected_columns = get_pattern_cols(columns, regex)
    selected_columns_idx = [columns.index(col) for col in selected_columns]
    selected_coef = coef[selected_columns_idx]
    component = x[selected_columns].dot(selected_coef)
    if include_intercept:
        component += intercept
    return component


def plot_change(
        observation=None,
        adaptive_lasso_estimate=None,
        trend_change=None,
        trend_estimate=None,
        year_seasonality_estimate=None,
        seasonality_change=None,
        seasonality_estimate=None,
        title=None,
        xaxis="Dates",
        yaxis="Values"):
    """Makes a plot of the observed data and estimated components, as well as detected changes

    The function currently allows five different components to be plotted together. Specifically,
    ``trend_change`` can be plotted with at least one of ``observations``, ``trend_estimate`` and
    ``adaptive_lasso_estimate`` is provided.

    Parameters
    ----------
    observation : `pandas.Series` with time index or `None`
        The observed values, leave None to omit it from plot.
    adaptive_lasso_estimate : `pandas.Series` with time index or `None`
        The adaptive lasso estimated trend, leave None to omit it from plot.
    trend_change : `pandas.Series` with time index or `None`
        The detected trend change points, leave None to omit it from plot.
        Plotted as vertical lines if observation is provided, otherwise plotted as markers.
    trend_estimate : `pandas.Series` with time index or `None`
        The estimated trend, leave None to omit it from plot.
    year_seasonality_estimate : `pandas.Series` with time index or `None`
        The estimated yearly seasonality, leave None to omit it from plot.
    seasonality_change : `list` or `dict` or `None`
        The detected seasonality change points, leave None to omit it from plot.
        If the type is `list`, it should be a list of change points of all components.
        If the type is `dict`, its keys should be the name of components, and values
        should be the corresponding list of change points.
    seasonality_estimate : `pandas.Series` or `None`
        The estimated seasonality, leave None to omit it from plot.
    title : `str` or `None`
        Plot title.
    xaxis : `str`
        Plot x axis label.
    yaxis : `str`
        Plot y axis label.

    Returns
    -------
    fig : `plotly.graph_objects`
        The plotted plotly object, can be shown with `iplot(fig)`.
    """
    if title is None:
        if trend_change is None and seasonality_change is None:
            title = "Timeseries Plot"
        elif trend_change is None:
            title = "Timeseries Plot with detected seasonality change points"
        elif seasonality_change is None:
            title = "Timeseries Plot with detected trend change points"
        else:
            title = "Timeseries Plot with detected trend and seasonality change points"
    fig = go.Figure()
    # shows the true observation
    if observation is not None:
        fig.add_trace(
            go.Scatter(
                x=observation.index,
                y=observation,
                name="true",
                mode="lines",
                line=dict(color="#42A5F5"),  # blue 400
                opacity=1)
        )
    # shows the seasonality estimate
    if seasonality_estimate is not None:
        fig.add_trace(
            go.Scatter(
                name="seasonality+trend",
                mode="lines",
                x=seasonality_estimate.index,
                y=seasonality_estimate,
                line=dict(color="#B2FF59"),  # light green A200
                opacity=0.7,
                showlegend=True)
        )
    # shows the adaptive lasso estimated trend
    if adaptive_lasso_estimate is not None:
        fig.add_trace(
            go.Scatter(
                name="adaptive lasso estimated trend",
                mode="lines",
                x=adaptive_lasso_estimate.index,
                y=adaptive_lasso_estimate,
                line=dict(color="#FFA726"),  # orange 400
                opacity=1,
                showlegend=True)
        )
    # shows the detected trend change points
    # shown as vertical lines if ``observation`` is provided, otherwise as markers
    if trend_change is not None:
        if observation is not None or seasonality_estimate is not None:
            if observation is not None:
                min_y = min(0, observation.min())
                max_y = max(0, observation.max())
            else:
                min_y = min(0, seasonality_estimate.min())
                max_y = max(0, seasonality_estimate.max())
            for i, cp in enumerate(trend_change):
                showlegend = (i == 0)
                fig.add_trace(
                    go.Scatter(
                        name="trend change point",
                        mode="lines",
                        x=[pd.to_datetime(cp), pd.to_datetime(cp)],
                        y=[min_y, max_y],
                        line=go.scatter.Line(
                            color="#F44336",  # red 500
                            width=1.5,
                            dash="dash"
                        ),
                        opacity=1,
                        showlegend=showlegend,
                        legendgroup="trend"
                    )
                )
        elif trend_estimate is not None or adaptive_lasso_estimate is not None:
            trend = trend_estimate if trend_estimate is not None else adaptive_lasso_estimate
            for cp in trend_change:
                value = trend[cp]
                fig.add_trace(
                    go.Scatter(
                        name="trend change points",
                        mode="markers",
                        x=[pd.to_datetime(cp)],
                        y=[value],
                        marker=dict(
                            color="#F44336",  # red 500
                            size=15,
                            line=dict(
                                color="Black",
                                width=1.5
                            )),
                        opacity=0.2,
                        showlegend=False
                    )
                )
        else:
            warnings.warn("trend_change is not shown. Must provide observations, trend_estimate, "
                          "adaptive_lasso_estimate or seasonality_estimate to plot trend_change.")
            return None
    # shows the detected seasonality change points
    if seasonality_change is not None:
        if observation is not None or seasonality_estimate is not None:
            if observation is not None:
                min_y = min(0, observation.min())
                max_y = max(0, observation.max())
            else:
                min_y = min(0, seasonality_estimate.min())
                max_y = max(0, seasonality_estimate.max())
            if isinstance(seasonality_change, list):
                for i, cp in enumerate(seasonality_change):
                    showlegend = (i == 0)
                    fig.add_trace(
                        go.Scatter(
                            name="seasonality change point",
                            mode="lines",
                            x=[pd.to_datetime(cp), pd.to_datetime(cp)],
                            y=[min_y, max_y],
                            line=go.scatter.Line(
                                color="#1E88E5",  # blue 600
                                width=1.5,
                                dash="dash"
                            ),
                            opacity=1,
                            showlegend=showlegend,
                            legendgroup="seasonality"
                        )
                    )
            elif isinstance(seasonality_change, dict):
                colors = ["#2196F3", "#0D47A1"]  # [blue 500, blue 900]
                dashes = ["dash", "dot", "dashdot", None]
                types = [(color, dash) for color in colors for dash in dashes]
                quota = len(types)
                for i, (key, changepoints) in enumerate(seasonality_change.items()):
                    if quota == 0:
                        warnings.warn("Only the first 8 components with detected change points"
                                      "are plotted.")
                        break
                    if changepoints:
                        for j, cp in enumerate(changepoints):
                            showlegend = (j == 0)
                            fig.add_trace(
                                go.Scatter(
                                    name="seasonality change point " + key,
                                    mode="lines",
                                    x=[pd.to_datetime(cp), pd.to_datetime(cp)],
                                    y=[min_y, max_y],
                                    line=go.scatter.Line(
                                        color=types[i][0],
                                        width=1.5,
                                        dash=types[i][1]
                                    ),
                                    opacity=1,
                                    showlegend=showlegend,
                                    legendgroup=key
                                )
                            )
                        quota -= 1
            else:
                raise ValueError("seasonality_change must be either list or dict.")
        else:
            warnings.warn("seasonality_change is not shown. Must provide observations or"
                          " seasonality_estimate to plot seasonality_change.")
    # shows the estimated trend
    if trend_estimate is not None:
        fig.add_trace(
            go.Scatter(
                name="trend",
                mode="lines",
                x=trend_estimate.index,
                y=trend_estimate,
                line=dict(
                    color="#FFFF00",  # yellow A200
                    width=2),
                opacity=1,
                showlegend=True)
        )
    # shows the estimated yearly seasonality
    if year_seasonality_estimate is not None:
        fig.add_trace(
            go.Scatter(
                name="yearly seasonality",
                mode="lines",
                x=year_seasonality_estimate.index,
                y=year_seasonality_estimate,
                line=dict(color="#BF360C"),  # deep orange 900
                opacity=0.6,
                showlegend=True)
        )
    fig.layout.update({
        "xaxis": dict(title=xaxis),
        "yaxis": dict(title=yaxis),
        "title": title
    })
    return fig


def adaptive_lasso_cv(x,
                      y,
                      initial_coef,
                      regularization_strength=None,
                      max_min_ratio=1e6):
    """Performs the adaptive lasso cross-validation.

     Algorithm is based on a transformation of `sklearn.linear_model.LassoCV()`.
     If initial_coef is not available, a lasso estimator is computed.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    initial_coef : `str` in ["ridge", "lasso", "old"] or `numpy.array`
        How to obtain the initial estimator. If a `str` is provided, the corresponding model is trained
        to obtain the initial estimator. If a `numpy.array` is provided, it is used as the initial
        estimator.
    regularization_strength : `float` in [0, 1] or `None`
        The regularization strength for change points. Greater values imply fewer change points.
        0 indicates all change points, and 1 indicates no change point.
        If `None`, cross-validation will be used to select tuning parameter,
        else the value will be used as the tuning parameters.
    max_min_ratio : `float`
        defines the min lambda by defining the ratio of lambda_max / lambda_min.
        `sklearn.linear_model.lasso_path` uses 1e3, but 1e6 seems better here.

    Returns
    -------
    intercept : `float`
        The estimated intercept.
    coef : `numpy.array`
        The estimated coefficients.
    """
    if regularization_strength is not None and (regularization_strength < 0 or regularization_strength > 1):
        raise ValueError("regularization_strength must be between 0.0 and 1.0.")
    if regularization_strength == 0:
        # regularization_strength == 0 implies linear regression
        model = LinearRegression().fit(x, y)
        return model.intercept_, model.coef_
    if regularization_strength == 1:
        # regularization_strength == 1 implies no change point selected
        # handle this case here separately for algorithm convergence and rounding concerns
        intercept = y.mean()
        coef = np.zeros(x.shape[1])
        return intercept, coef
    if type(initial_coef) == str:
        model = {"lasso": LassoCV(), "ols": LinearRegression(), "ridge": RidgeCV()}[initial_coef]
        model.fit(x, y)
        initial_coef = model.coef_
    else:
        assert x.shape[1] == len(initial_coef), (
            "the number of columns in x should equal to the length of weights"
            f"but got {x.shape[1]} and {len(initial_coef)}."
        )
    weights = np.where(initial_coef != 0, 1 / abs(initial_coef), 1e16)  # 1e16 is big enough for most cases.
    x_t = x / weights
    if regularization_strength is None:
        model = LassoCV().fit(x_t, y)
    else:
        # finds the minimum lambda that corresponds to no selection, formula derived from KKT condition.
        # this is the max lambda we need to consider
        max_lam = max(abs(np.matmul(x_t.T, y - y.mean()))) / x.shape[0]
        # the lambda we choose is ``regularization_strength``th log-percentile of [lambda_min, lambda_max]
        lam = 10 ** (np.log10(max_lam / max_min_ratio) + np.log10(max_min_ratio) * regularization_strength)
        model = Lasso(alpha=lam).fit(x_t, y)
    intercept = model.intercept_
    coef = model.coef_ / weights
    return intercept, coef


def find_neighbor_changepoints(cp_idx, min_index_distance=2):
    """Finds neighbor change points given their indices

    For example
        x = [1, 2, 3, 7, 8, 10, 15, 20]
    The return with `min_index_distance`=2 is
        result = [[1, 2, 3], [7, 8], [10], [15], [20]]
    The return with `min_index_distance`=3 is
        result = [[1, 2, 3], [7, 8, 10], [15], [20]]

    Parameters
    ----------
    cp_idx : `list`
        A list of the change point indices.
    min_index_distance : `int`
        The minimal index distance to be considered separate

    Returns
    -------
    neighbor_cps : `list`
        A list of neighbor change points lists. Single change points are put in a list as well.
    """
    if min_index_distance <= 0:
        raise ValueError("`min_index_distance` must be positive.")
    if len(cp_idx) <= 1:
        return [cp_idx]
    # check if sorted
    is_sorted = True
    for i in range(1, len(cp_idx)):
        if cp_idx[i] < cp_idx[i - 1]:
            is_sorted = False
            break
    # sort if `cp_idx` is not sorted
    if not is_sorted:
        cp_idx.sort()
        warnings.warn("The given `cp_idx` is not sorted. It has been sorted.")
    neighbor_cps = []
    i = 0
    while i < len(cp_idx):
        neighbor_cps.append([])
        neighbor_cps[-1].append(cp_idx[i])
        i += 1
        while i < len(cp_idx) and cp_idx[i] < neighbor_cps[-1][-1] + min_index_distance:
            neighbor_cps[-1].append(cp_idx[i])
            i += 1
    return neighbor_cps


def get_trend_changes_from_adaptive_lasso(x,
                                          y,
                                          changepoint_dates,
                                          initial_coef,
                                          min_index_distance=2,
                                          regularization_strength=None):
    """Parses the adaptive lasso estimator to get change point dates.

    The functions calls ``adaptive_lasso_cv`` to selected potential trend change points.
    Then a filter is applied to eliminate change points that are too close.
    Specifically, in a set of close change points, the one with the largest absolute
    coefficient will be kept.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    changepoint_dates : `pandas.Series`
        A pandas Series of all potential change point dates that were used to generate ``x``.
    initial_coef : `str` in ["ridge", "lasso", "old"] or `numpy.array'
        How to obtain the initial estimator. If a `str` is provided, the corresponding model is trained
        to obtain the initial estimator. If a `numpy.array` is provided, it is used as the initial
        estimator.
    min_index_distance : `int`
        The minimal index distance that is allowed between two change points.
    regularization_strength : `float` in [0, 1] or `None`
        The regularization power for change points. Greater values imply fewer change points.
        0 indicates all change points, and 1 indicates no change point.
        If `None`, cross-validation will be used to select tuning parameter,
        else the value will be used as the tuning parameters.

    Returns
    -------
    changepoints : `list`
        Detected trend change points.
    coefs : `list`
        Adaptive lasso estimated coefficients.
        First element is intercept, and second element is coefficients.
    """
    intercept, coef = adaptive_lasso_cv(x, y, initial_coef, regularization_strength)
    # x may contain yearly seasonality regressors, so we need i < len(changepoint_dates)
    nonzero_idx = [i for i, val in enumerate(coef) if val != 0 and i < len(changepoint_dates)]
    cp_blocks = find_neighbor_changepoints(
        cp_idx=nonzero_idx,
        min_index_distance=min_index_distance,
    )
    cp_idx = filter_changepoints(cp_blocks, coef, min_index_distance)
    return changepoint_dates.iloc[cp_idx].tolist(), [intercept, coef]


def compute_min_changepoint_index_distance(
        df,
        time_col,
        n_changepoints,
        min_distance_between_changepoints):
    """Computes the minimal index distance between two consecutive detected change points.

    Given a df, its time column, the number of change points that are evenly distributed,
    and the min_distance_between_changepoints in `DateOffset, Timedelta or str`, gets the
    min distance between change point indices.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The dataframe that has a time column.
    time_col : `str`
        The column name of time in ``df``.
    n_changepoints : `int`
        Number of change points that are uniformly placed over the time period.
    min_distance_between_changepoints : `DateOffset`, `Timedelta` or `str`
        The minimal distance that is allowed between two detected change points.
        Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
        '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
        cycles or indefinite number of days, thus is not parsable by pandas as timedelta.

    Returns
    -------
    min_changepoint_index_distance : `int`
        The minimal index distance that is allowed between two detected change points.
    """
    if n_changepoints == 0:
        return df.shape[0]
    check_freq_unit_at_most_day(min_distance_between_changepoints, "min_distance_between_changepoints")
    # `to_offset` function handles string frequency without numbers such as 'D' instead of '1D',
    # so we use `to_offset` here rather than `to_timedelta`.
    min_dist = to_offset(min_distance_between_changepoints)
    # There are `n_changepoints` change points and 1 growth term, so there is `n_changepoints` gaps
    # Therefore the following is divided by `n_changepoints`
    try:
        changepoint_dist = (df[time_col].iloc[-1] - df[time_col].iloc[0]) / n_changepoints
    except TypeError:
        changepoint_dist = (pd.to_datetime(df[time_col].iloc[-1]) - pd.to_datetime(df[time_col].iloc[0])) \
                           / n_changepoints
    return int(np.ceil(min_dist.delta.total_seconds() / changepoint_dist.total_seconds()))


def filter_changepoints(cp_blocks, coef, min_index_distance):
    """Filters change points that are too close.

    Given the ``cp_block`` from the output of ``find_neighbor_changepoints`` and the corresponding
    coefficients, finds the list of change points, one in each neighborhood list.

    The algorithm keeps all individual change points. If more than one change points are in the
    same block, with the maximum distance less than ``min_index_distance``, then only the one
    with the maximum absolute coefficient is retained. If more than one change points are in the
    same block, and the block covers a few ``max_index_distance`` length, then a greedy method
    is used to select change points. The principle is that, we perform one pass of the change
    points, within each ``max_index_distance``, we select one change point with the maximum
    absolute coefficients. If the next change point is with in ``max_index_distance`` of the
    previous one, the previous one is dropped. A back-tracking is also used to fill possible
    change points when a change point is dropped.

    Parameters
    ----------
    cp_blocks : `list`
        A list of list of change points, output from ``find_neighbor_changepoints``.
    coef : `numpy.array`
        The estimated coefficients for all potential change points.
        Note: ``coef`` is a 1-D array, which is the output from regression model.
    min_index_distance : `int`
        The minimum index distance between two selected change points.
        Note that this parameter is added to safe guard the extreme cases: if all
        significant change points are included in the same block, with insignificant
        change points connecting them, then we do not want to drop all and just leave
        one change points. With this parameter, we can leave more than one change point
        from each block, and keep them at least ``min_index_distance`` away.

    Returns
    -------
    changepoint_indices : `list`
        The indices of selected change points.
    """
    if min_index_distance <= 0:
        raise ValueError("`min_index_distance` is the minimum distance between change point"
                         "indices to consider them separate, and must be positive.")
    if min_index_distance == 1:
        return [cp for block in cp_blocks for cp in block]
    coef = coef.ravel()  # makes sure ``coef`` is 1-D array
    coef = coef[[i for block in cp_blocks for i in block]]  # Only needs the coefs that correspond to selected cp's.
    selected_changepoints = []
    if cp_blocks == [[]]:
        return []
    for i, block in enumerate(cp_blocks):
        if len(block) == 1:
            # if only one cp in a block, leaves it
            selected_changepoints.append(block[0])
        else:
            # needs coefs to decide
            coef_start = sum([len(block) for block in cp_blocks[:i]])
            coef_block = coef[coef_start: coef_start + len(block)]
            if block[-1] - block[0] < min_index_distance:
                # If block max distance is less than `min_index_distance`,
                # only one cp can be selected,
                # select the one with the greatest absolute coef
                selected_changepoints.append(block[abs(coef_block).argmax()])
            else:
                # For longer blocks, the change points with the maximum
                # absolute coefficients are selected, while keeping the
                # distance beyond ``min_index_distance``.
                # This prevents dropping too many change points in a long block.
                block_cps = []
                start = block[0]
                last_coef = 0
                while start <= block[-1]:
                    # gets the current ``min_index_distance`` sized sub-block of cps and coefs
                    current_sub_block = [cp for cp in block if start <= cp < start + min_index_distance]
                    if not current_sub_block:
                        start += min_index_distance
                        continue
                    current_cp = current_sub_block[
                        abs(coef_block[[block.index(cp) for cp in current_sub_block]]).argmax()]
                    current_cp_idx = block.index(current_cp)
                    current_coef = coef_block[current_cp_idx]
                    # the first change point, doesn't have to look back
                    if start == block[0]:
                        block_cps.append(current_cp)
                    else:
                        # check the distance from the last change point
                        # if distance is enough, we can fit the new change point
                        last_cp = block_cps[-1]
                        if current_cp - last_cp >= min_index_distance:
                            block_cps.append(current_cp)
                        # if the distance is not enough and the new cp has a greater absolute coef,
                        # we have to remove the previous one with less absolute coef
                        # however, we would like to see if another cp fits in between
                        elif abs(current_coef) > abs(last_coef):
                            # checks if an extra change point can fit between last_last_cp and current_cp
                            if len(block_cps) >= 2:
                                last_last_cp = block_cps[-2]
                            else:
                                # if no last_last_cp, there's not lower bound
                                last_last_cp = -min_index_distance
                            # gets the cps that can fit between last_last_cp and current_cp
                            back_fill_potential_cp = [
                                cp for cp in block if
                                last_last_cp + min_index_distance <= cp <= current_cp - min_index_distance]
                            if back_fill_potential_cp:
                                # if there is cp that can fit in between, gets the one with max absolute coef
                                potential_coef = coef[[block.index(cp) for cp in back_fill_potential_cp]]
                                back_fill_cp = back_fill_potential_cp[abs(potential_coef).argmax()]
                                # append the cp in between
                                block_cps.append(back_fill_cp)
                            # removes the last one and append the new one
                            block_cps.remove(last_cp)
                            block_cps.append(current_cp)
                    # updates last_coef for comparison
                    last_coef = current_coef
                    # search ahead
                    start += min_index_distance
                selected_changepoints += block_cps
    return selected_changepoints


def get_changes_from_beta(
        beta,
        seasonality_components_df,
        magnitude_only=False):
    """Gets the seasonality magnitude change arrays for each seasonality component from the
    estimated regression coefficients.

    In total there are::

        {sum_{i=1}^{number of components}(2 * order of component i)} * {number of changepoints + 1}

    coefficients in ``beta``, where the + 1 counts the overall seasonality. These coefficients
    indicate the change in magnitude for each term in each component at each change point.
    We can't work on these directly, since the values are positive or negative.

    Two options are available to calcumated the change metric, controlled by the parameter
    ``magnitude_only``.

    If ``magnitude_only`` == True, the cumulative sum is taken for every single cos or sin term
    (from all components) along the change points. The results are the magnitudes for each cos
    or sin term (from all components) at each change point. An L2 norm is taken over the cos and sin
    terms' magnitudes within the same component at the same change point. Now for each component
    at every change point, we have one L2 norm value that represents the magnitude (no longer
    cos or sin level). Then the magnitude changes for the same component between consecutive
    change points are computed. This option captures total magnitude changes only.

    If ``magnitude_only`` == False, an L2 norm is taken over the cos or sin magnitude changes
    in the same component directly. This option captures shape changes as well.

    A dictionary is returned with keys equal to the component names, and values are the corresponding
    arrays of changes.

    Parameters
    ----------
    beta : `np.array`
        The estimated regression coefficients.
    seasonality_components_df : `pandas.DataFrame`
        The df to generate seasonality design matrix, which is compatible with
        ``seasonality_components_df`` in
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`
    magnitude_only : `bool`, default False
        Set to True to compute the L2 norms on the seasonality magnitudes, and set to False to compute the
        L2 norms on the seasonailty changes.

    Returns
    -------
    seasonality_magnitude : `dict`
        Keys are seasonality components, and values are the arrays of changes at each changepoint for the
        components.

    Examples
    --------
    # 1 change point, 2 seasonality components: "weekly" has order 1, and "yearly" has order 2.
    # so (1 + 2) * 2 * (1 + 1) = 12 coefficients in total.
    # the terms above are
    # [(weekly_order + yearly_order) * 2_terms_sin_cos] * (overall_seasonality + num_changepoints)
    >>> # the initial coefficients are [1, 1, 1, 1, 1, 1], and changes to [2, 0, 2, 2, 2, 0] at the change point
    >>> # the change is [1, -1, 1, 1, 1, -1].
    >>> beta = np.array([1, 1,     1, 1, 1, 1,          # initial coefficients
    ...                  1, -1,    1, 1, 1, -1])        # changes to [2, 0,     2, 2, 2, 0]
    >>> seasonality_components_df = pd.DataFrame({
    ...     "name": ["tow", "conti_year"],
    ...     "period": [7.0, 1.0],
    ...     "order": [1, 2],
    ...     "seas_names": ["weekly", "yearly"]})
    >>> result = get_changes_from_beta(beta, seasonality_components_df, True)
    >>> # when ``magnitude_only`` is set to True, the norm is calculated on the magnitude coefficients
    >>> # we have norms = [[sqrt(2), 2], [2, 2 * sqrt(3)]]
    >>> # hence the changes at the change point are [2 - sqrt(2), 2 * sqrt(3) - 2]
    >>> # the first coefficients is always prepended.
    >>> for key, value in result.items():
    >>>     print(key, value)
    weekly [1.41421356 0.58578644]
    yearly [2.         1.46410162]

    >>> # when ``magnitude_only`` is set to False, the norm is calculated directly on the change coefficients
    >>> # we have changes [sqrt(2), 2] at the change point
    >>> result = get_changes_from_beta(beta, seasonality_components_df, False)
    >>> for key, value in result.items():
    >>>     print(key, value)
    weekly [1.41421356 1.41421356]
    yearly [2. 2.]
    """
    # gets the number of terms in each component
    num_terms = (seasonality_components_df["order"] * 2).tolist()
    # gets the names of components
    components = seasonality_components_df["seas_names"].tolist()
    # reshapes the regression coefficients and computes the magnitude coefficients
    # the columns are the terms of different orders for each component
    # each row is one change point
    beta_mat = beta.reshape(-1, sum(num_terms))
    if magnitude_only:
        # cumsum gets the coefficient for the term between this changepoint and the next one.
        coef_mat = np.cumsum(beta_mat, axis=0)
    # computes the change magnitudes of each component at each time point, metric is l2 norm
    result = {}
    for i in range(len(num_terms)):
        start = np.sum(num_terms[:i]).astype(int)
        end = np.sum(num_terms[:(i + 1)]).astype(int)
        if magnitude_only:
            result[components[i]] = np.diff(np.linalg.norm(coef_mat[:, start: end], axis=1), prepend=0)
        else:
            result[components[i]] = np.linalg.norm(beta_mat[:, start: end], axis=1)
    return result


def get_seasonality_changes_from_adaptive_lasso(
        x,
        y,
        changepoint_dates,
        initial_coef,
        seasonality_components_df,
        min_index_distance=2,
        regularization_strength=0.6):
    """Parses the adaptive lasso estimator to get change point dates.

    The functions calls ``adaptive_lasso_cv`` to selected potential seasonality change points.
    Then a filter is applied to eliminate change points that are too close.
    Specifically, in a set of close change points, the one with the largest absolute
    coefficient will be kept.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    changepoint_dates : `pandas.Series`
        A pandas Series of all potential change point dates that were used to generate ``x``.
    initial_coef : `str` in ["ridge", "lasso", "old"] or `numpy.array'
        How to obtain the initial estimator. If a `str` is provided, the corresponding model is trained
        to obtain the initial estimator. If a `numpy.array` is provided, it is used as the initial
        estimator.
    seasonality_components_df : `pandas.DataFrame`
        The df to generate seasonality design matrix, which is compatible with
        ``seasonality_components_df`` in
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`
    min_index_distance : `int`
        The minimal index distance that is allowed between two change points.
    regularization_strength : `float` in [0, 1]
        The regularization power for change points. Greater values imply fewer change points.
        0 indicates all change points, and 1 indicates no change point.

    Returns
    -------
    result : `dict`
        The detected seasonality change points result dictionary. Keys are the component names,
        and values are the corresponding detected change points.
    """
    intercept, beta = adaptive_lasso_cv(
        x=x,
        y=y,
        initial_coef=initial_coef,
        regularization_strength=regularization_strength,
        # Typically, seasonality has fewer changepoints than trend. Lower ratio results in higher lambda.
        max_min_ratio=1e4)
    change_result = get_changes_from_beta(
        beta=beta,
        seasonality_components_df=seasonality_components_df)
    result = dict()
    for component, changes in change_result.items():
        nonzero_idx = [i for i, val in enumerate(changes) if val != 0]
        cp_blocks = find_neighbor_changepoints(
            cp_idx=nonzero_idx,
            min_index_distance=min_index_distance,
        )
        cp_idx = filter_changepoints(cp_blocks, changes, min_index_distance)
        result[component] = changepoint_dates.iloc[cp_idx].tolist()
    return result


def estimate_trend_with_detected_changepoints(
        df,
        time_col,
        value_col,
        changepoints,
        yearly_seasonality_order=8,
        estimator="ridge"):
    """Estimates the trend effect with detected change points.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data df.
    time_col : `str`
        The column name of the time column in ``df``.
    value_col : `str`
        The column name of the value column in ``df``.
    changepoints : `list`
        A list of detected trend change points.
    yearly_seasonality_order : `int`
        The yearly seasonality order.
    estimator : `str`, default "ridge"
        "ols" or "ridge", the estimation model for trend estimation.

    Returns
    -------
    trend_estimate : `pandas.Series`
        The estimated trend.
    """
    trend_df = build_trend_feature_df_with_changes(
        df=df,
        time_col=time_col,
        changepoints_dict={
            "method": "custom",
            "dates": changepoints
        }
    )
    if yearly_seasonality_order > 0:
        long_seasonality_df = build_seasonality_feature_df_with_changes(
            df=df,
            time_col=time_col,
            fs_components_df=pd.DataFrame({
                "name": ["conti_year"],
                "period": [1.0],
                "order": [yearly_seasonality_order],
                "seas_names": ["yearly"]})
        )
        trend_df = pd.concat([trend_df, long_seasonality_df], axis=1)
    estimators = {
        "ridge": RidgeCV,
        "ols": LinearRegression
    }
    estimator = estimators.get(estimator)
    if estimator is None:
        raise ValueError("estimator can only be either 'ridge' or 'ols'.")
    non_na_index = ~df[value_col].isnull().values
    model = estimator().fit(trend_df.values[non_na_index], df[value_col].values[non_na_index])
    # can't simply do .predict, because need to exclude the seasonality terms.
    trend_estimate = np.matmul(
        trend_df.values[:, :(len(changepoints) + 1)], model.coef_[:(len(changepoints) + 1)]) \
        + model.intercept_
    trend_estimate = pd.Series(trend_estimate)
    trend_estimate.index = pd.to_datetime(df[time_col])
    return trend_estimate


def estimate_seasonality_with_detected_changepoints(
        df,
        time_col,
        value_col,
        seasonality_changepoints,
        seasonality_components_df,
        estimator="ols"):
    """Estimates the seasonality effect with detected change points.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data df.
    time_col : `str`
        The column name of the time column in ``df``.
    value_col : `str`
        The column name of the value column in ``df``.
    seasonality_changepoints : `dict`
        The detected seasonality change points dictionary, output from
        `~greykite.algo.changepoint.adalasso.changepoints_utils.get_seasonality_changes_from_adaptive_lasso`
    seasonality_components_df : `pandas.DataFrame`
        The df to generate seasonality design matrix, which is compatible with
        ``seasonality_components_df`` in
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`
    estimator : `str`, default "ols"
        "ols" or "ridge", the estimation model for seasonality estimation.

    Returns
    -------
    seasonality_estimate : `pandas.Series`
        The estimated seasonality.
    """
    seasonality_df = build_seasonality_feature_df_from_detection_result(
        df=df,
        time_col=time_col,
        seasonality_changepoints=seasonality_changepoints,
        seasonality_components_df=seasonality_components_df
    )
    estimators = {
        "ridge": RidgeCV,
        "ols": LinearRegression
    }
    estimator = estimators.get(estimator)
    if estimator is None:
        raise ValueError("estimator can only be either 'ridge' or 'ols'.")
    non_na_index = ~df[value_col].isnull().values
    model = estimator().fit(seasonality_df.values[non_na_index], df[value_col].values[non_na_index])
    seasonality_estimate = model.predict(seasonality_df.values)
    seasonality_estimate = pd.Series(seasonality_estimate)
    seasonality_estimate.index = pd.to_datetime(df[time_col])
    return seasonality_estimate


def get_seasonality_changepoint_df_cols(
        df,
        time_col,
        seasonality_changepoints,
        seasonality_components_df,
        include_original_block=True,
        include_components=None):
    """Gets the seasonality change point feature df column names.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The dataframe used to build seasonality feature df.
    time_col : `str`
        The name of the time column in ``df``.
    seasonality_changepoints : `dict`
        The seasonality change point dictionary. The keys are seasonality components, and the values
        are the corresponding change points given in lists.
        For example

            "weekly": [Timestamp('2020-01-01 00:00:00'), Timestamp('2021-04-05 00:00:00')]
            "yearly": [Timestamp('2020-08-06 00:00:00')]

    seasonality_components_df : `pandas.DataFrame`
        The seasonality components dataframe that is compatible with
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`
        The values in "seas_names" must equal to the keys in ``seasonality_changepoints``.
    include_original_block : `bool`, default True
        Whether to include the original untruncated block of sin or cos columns for each component. If set to
        False, the original seasonality block for each component will be dropped.
    include_components : `list` [`str`] or None, default None
        The components to be included from the result. If None, all components will be included.

    Returns
    -------
    cols : `list`
        List of column names for seasonality change points df.
    """
    # only needs two rows to get the column names, reduces running time.
    df = df.iloc[:2, :]
    seasonality_cols = []
    for component, dates in seasonality_changepoints.items():
        if include_components is None or component in include_components:
            regular_seasonality_df = build_seasonality_feature_df_with_changes(
                df=df,
                time_col=time_col,
                fs_components_df=seasonality_components_df.loc[seasonality_components_df["seas_names"] == component, :]
            )
            temp_cols = list(regular_seasonality_df.columns) if include_original_block else []
            for date in dates:
                temp_cols += [f"{col}{date.strftime('_%Y_%m_%d_%H')}" for col in list(regular_seasonality_df.columns)]
            seasonality_cols += temp_cols
    return seasonality_cols


def get_trend_changepoint_dates_from_cols(trend_cols):
    """Gets the trend changepoint dates from trend changepoint column names.

    Parameters
    ----------
    trend_cols : `list[`str`]`
        List of trend changepoint column names. EX. "changepoint2_2018_01_05_00".

    Returns
    -------
    trend_changepoint_dates : `list[`timestamp`]`
        List of trend changepoint dates.
    """
    trend_changepoint_dates = []
    trend_cols = get_pattern_cols(trend_cols, f"^{CHANGEPOINT_COL_PREFIX}")
    if trend_cols:
        for col in trend_cols:
            date = col.split("_")[1:]
            date = date + ['00'] * (6 - len(date))  # 6 means Y m d H M S
            date = f'{"-".join(date[:3])} {":".join(date[3:])}'  # Y-m-d H:M:S
            trend_changepoint_dates.append(pd.to_datetime(date))
    return trend_changepoint_dates


def get_yearly_seasonality_changepoint_dates_from_freq(
        df,
        time_col,
        yearly_seasonality_change_freq,
        min_training_length="183D"):
    """Gets the yearly seasonality changepoint dates from change frequency.

    This is an internal function used for varying yearly seasonality effects in
    `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_trend_changepoints`
    For a given ``yearly_seasonality_change_freq``, for example "365D", it generates changepoint
    dates according to this frequency within the timeframe of ``df``. The length of data after the last
    changepoint date can not be less than ``least_training_length``.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data df.
    time_col : `str`
        The column name for time column in ``df``.
    yearly_seasonality_change_freq : `DateOffset`, `Timedelta` or `str` or `None`
        How often to change the yearly seasonality model, i.e., how often to place yearly seasonality
        changepoints.
        Note that if you use `str` as input, the maximal supported unit is day, i.e.,
        you might use "200D" but not "12M" or "1Y".
    min_training_length : `DateOffset`, `Timedelta` or `str` or `None`, default "183D"
        The minimum length between the last changepoint date and the last date of ``df[time_col]``.
        Changepoints too close to the end will be omitted. Recommended at least half a year.
        Note that if you use `str` as input, the maximal supported unit is day, i.e.,
        you might use "200D" but not "12M" or "1Y".

    Returns
    -------
    yearly_seasonality_changepoint_dates : `list` [ `pandas._libs.tslibs.timestamps.Timestamp` ]
        A list of yearly seasonality changepoint dates.
    """
    if yearly_seasonality_change_freq is None:
        return []
    check_freq_unit_at_most_day(yearly_seasonality_change_freq, "yearly_seasonality_change_freq")
    yearly_seasonality_change_freq = to_offset(yearly_seasonality_change_freq)
    if yearly_seasonality_change_freq.delta < timedelta(days=365):
        warnings.warn("yearly_seasonality_change_freq is less than a year. It might be too short"
                      "to fit accurate yearly seasonality.")
    check_freq_unit_at_most_day(min_training_length, "least_training_length")
    if min_training_length is not None:
        min_training_length = to_offset(min_training_length)
    else:
        min_training_length = to_offset("0D")
    first_day_in_df = pd.to_datetime(df[time_col].iloc[0])
    last_day_in_df = pd.to_datetime(df[time_col].iloc[-1])
    yearly_seasonality_changepoint_dates = list(pd.date_range(
        start=first_day_in_df,
        end=last_day_in_df - min_training_length,
        freq=yearly_seasonality_change_freq))[1:]  # do not include the start date
    if len(yearly_seasonality_changepoint_dates) == 0:
        warnings.warn("No yearly seasonality changepoint added. Either data length is too short"
                      "or yearly_seasonality_change_freq is too long.")
    return yearly_seasonality_changepoint_dates


def combine_detected_and_custom_trend_changepoints(
        detected_changepoint_dates,
        custom_changepoint_dates,
        min_distance=None,
        keep_detected=False):
    """Adds custom trend changepoints to detected trend changepoints.

    Compares the distance between custom changepoints and detected changepoints,
    and drops a detected changepoint or a custom changepoint depending on ``keep_detected``
    if their distance is less than ``min_distance``.

    Parameters
    ----------
    detected_changepoint_dates : `list`
        A list of detected trend changepoints, parsable by `pandas.to_datetime`.
    custom_changepoint_dates : `list`
        A list of additional custom trend changepoints, parsable by `pandas.to_datetime`.
    min_distance : `DateOffset`, `Timedelta`, `str` or None, default None
        The minimum distance between detected changepoints and custom changepoints.
        If a detected changepoint and a custom changepoint have distance less than ``min_distance``,
        either the detected changepoint or the custom changepoint will be dropped according to ``keep_detected``.
        Does not compare the distance within detected changepoints or custom changepoints.
        Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
        '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
        cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        For example, see `pandas.tseries.frequencies.to_offset`.
    keep_detected : `bool`, default False
        When the distance of a detected changepoint and a custom changepoint is less than ``min_distance``,
        whether to keep the detected changepoint or the custom changepoint.

    Returns
    -------
    combined_changepoint_dates : `list`
        A list of combined changepoints in ascending order.
    """
    check_freq_unit_at_most_day(min_distance, "min_distance")
    detected_changepoint_dates = pd.to_datetime(detected_changepoint_dates)
    custom_changepoint_dates = pd.to_datetime(custom_changepoint_dates)
    if len(detected_changepoint_dates) == 0:
        return custom_changepoint_dates
    if len(custom_changepoint_dates) == 0:
        return detected_changepoint_dates
    if keep_detected:
        combined_changepoints = [cp for cp in detected_changepoint_dates]
        addition_changepoints = [cp for cp in custom_changepoint_dates]
    else:
        combined_changepoints = [cp for cp in custom_changepoint_dates]
        addition_changepoints = [cp for cp in detected_changepoint_dates]
    for addition_cp in addition_changepoints:
        if min_distance is not None:
            current_min_distance = min([abs(addition_cp - cp) for cp in combined_changepoints])
            if current_min_distance >= to_offset(min_distance):
                combined_changepoints.append(addition_cp)
        else:
            combined_changepoints.append(addition_cp)
    return sorted(combined_changepoints)
